"""
XPAK container support
"""

__all__ = ("MalformedXpak", "Xpak")

import os
from collections import OrderedDict

from snakeoil import klass
from snakeoil import struct_compat as struct

from ..exceptions import PkgcoreException

# format is:
# XPAKPACKIIIIDDDD[index][data]XPAKSTOPOOOOSTOP
# first; all ints/longs are big endian
# meanwhile, 8 byte format magic
# 4 bytes of index len,
# 4 bytes of data len
# index items: 4 bytes (len of the key name), then that length of key data
#   finally, 2 longs; relative offset from data block start, length of the data
#   repeats till index is full processed
# for data, just a big blob; offsets into it are determined via the index
#   table.
# finally, trailing magic, 4 bytes (positive) of the # of bytes to seek to
#   reach the end of the magic, and 'STOP'.  offset is relative to EOS for Xpak


class MalformedXpak(PkgcoreException):

    def __init__(self, msg):
        super().__init__(f"xpak as malformed: {msg}")
        self.msg = msg


class Xpak:
    __slots__ = ("_source", "_source_is_path", "xpak_start", "_keys_dict")

    _reading_key_rewrites = {'repo': 'REPO'}

    trailer_pre_magic = "XPAKSTOP"
    trailer_post_magic = "STOP"
    trailer = struct.Struct(">%isL%is" % (
        len(trailer_pre_magic), len(trailer_post_magic)))

    header_pre_magic = "XPAKPACK"
    header = struct.Struct(">%isLL" % (len(header_pre_magic),))

    trailer_post_magic = trailer_post_magic.encode("ascii")
    trailer_pre_magic = trailer_pre_magic.encode("ascii")
    header_pre_magic = header_pre_magic.encode("ascii")

    def __init__(self, source):
        self._source_is_path = isinstance(source, str)
        self._source = source
        self.xpak_start = None
        # keys_dict becomes an ordereddict after _load_offsets; reason for
        # it is so that reads are serialized.

    @property
    def _fd(self):
        # we do this annoying little dance to avoid having a couple
        # hundred fds open if they're accessing a lot of binpkgs
        if self._source_is_path:
            return open(self._source, "rb")
        return self._source

    @classmethod
    def write_xpak(cls, target_source, data):
        """
        write an xpak dict to disk; overwriting an xpak if it exists

        :param target_source: string path, or \
          :obj:`snakeoil.data_source.base` derivative
        :param data: mapping instance to write into the xpak.
        :return: xpak instance
        """
        try:
            old_xpak = cls(target_source)
            # force access
            list(old_xpak.keys())
            start = old_xpak.xpak_start
            source_is_path = old_xpak._source_is_path
        except (MalformedXpak, IOError):
            source_is_path = isinstance(target_source, str)
            if source_is_path:
                try:
                    start = os.lstat(target_source).st_size
                except FileNotFoundError:
                    start = 0
            else:
                f = target_source.bytes_fileobj(writable=True)
                f.seek(0, 2)
                start = f.tell()
        new_index = []
        new_data = []
        cur_pos = 0
        for key, val in data.items():
            if isinstance(val, str):
                val = val.encode('utf8')
            if isinstance(key, str):
                key = key.encode()
            new_index.append(struct.pack(
                ">L%isLL" % len(key),
                len(key), key, cur_pos, len(val)))
            new_data.append(val)
            cur_pos += len(val)

        if source_is_path:
            # rb+ required since A) binary, B) w truncates from the getgo
            handle = open(target_source, "r+b")
        else:
            handle = target_source.bytes_fileobj(writable=True)

        joiner = b''
        new_index = joiner.join(new_index)
        new_data = joiner.join(new_data)

        handle.seek(start, 0)
        cls.header.write(
            handle, cls.header_pre_magic, len(new_index), len(new_data))

        handle.write(struct.pack(
            ">%is%is" % (len(new_index), len(new_data)), new_index, new_data))

        # the +8 is for the longs for new_index/new_data
        cls.trailer.write(
            handle, cls.trailer_pre_magic,
            len(new_index) + len(new_data) + cls.trailer.size + 8,
            cls.trailer_post_magic)
        handle.truncate()
        handle.close()
        return Xpak(target_source)

    @klass.jit_attr
    def keys_dict(self):
        fd = self._fd
        index_start, index_len, data_len = self._check_magic(fd)
        data_start = index_start + index_len
        keys_dict = OrderedDict()
        key_rewrite = self._reading_key_rewrites.get
        while index_len:
            key_len = struct.unpack(">L", fd.read(4))[0]
            key = fd.read(key_len)
            key = key.decode('ascii')
            if len(key) != key_len:
                raise MalformedXpak(
                    "tried reading key %i of len %i, but hit EOF" % (
                        len(keys_dict) + 1, key_len))
            try:
                offset, data_len = struct.unpack(">LL", fd.read(8))
            except struct.error as e:
                raise MalformedXpak(
                    "key %i, tried reading data offset/len but hit EOF" % (
                        len(keys_dict) + 1)) from e
            key = key_rewrite(key, key)
            keys_dict[key] = (
                data_start + offset, data_len,
                not key.startswith("environment"))
            index_len -= (key_len + 12) # 12 for key_len, offset, data_len longs

        return keys_dict

    def _check_magic(self, fd):
        fd.seek(-16, 2)
        try:
            pre, size, post = self.trailer.read(fd)
            if pre != self.trailer_pre_magic or post != self.trailer_post_magic:
                raise MalformedXpak(
                    "not an xpak segment, trailer didn't match: %r" % fd)
        except struct.error as e:
            raise MalformedXpak(
                "not an xpak segment, failed parsing trailer: %r" % fd) from e

        # this is a bit daft, but the format seems to intentionally
        # have an off by 8 in the offset address. presumably cause the
        # header was added after the fact, either way we go +8 to
        # check the header magic.
        fd.seek(-(size + 8), 2)
        self.xpak_start = fd.tell()
        try:
            pre, index_len, data_len = self.header.read(fd)
            if pre != self.header_pre_magic:
                raise MalformedXpak(
                    "not an xpak segment, header didn't match: %r" % fd)
        except struct.error as e:
            raise MalformedXpak(
                "not an xpak segment, failed parsing header: %r" % fd) from e

        return self.xpak_start + self.header.size, index_len, data_len

    def keys(self):
        return self.keys_dict.keys()

    def values(self):
        fd = self._fd
        return (self._get_data(fd, *v) for v in self.keys_dict.values())

    def items(self):
        # note that it's an OrderedDict, so this works.
        fd = self._fd
        return (
            (k, self._get_data(fd, *v))
            for k, v in self.keys_dict.items())

    def __len__(self):
        return len(self.keys_dict)

    def __contains__(self, key):
        return key in self.keys_dict

    def __bool__(self):
        return bool(self.keys_dict)

    def __iter__(self):
        return iter(self.keys_dict)

    def __getitem__(self, key):
        return self._get_data(self._fd, *self.keys_dict[key])

    def __delitem__(self, key):
        del self.keys_dict[key]

    def __setitem__(self, key, val):
        self.keys_dict[key] = val
        return val

    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default

    def pop(self, key, *a):
        # faster then the exception form...
        l = len(a)
        if l > 1:
            raise TypeError("pop accepts 1 or 2 args only")
        if key in self.keys_dict:
            o = self.keys_dict.pop(key)
        elif l:
            o = a[0]
        else:
            raise KeyError(key)
        return o

    def _get_data(self, fd, offset, data_len, needs_decoding=False):
        # optimization for file objs; they cache tell position, but
        # pass through all seek calls (nice, eh?) so we rely on that
        # for cutting down on uneeded seeks; userland comparison being
        # far cheaper then an actual syscall seek
        if fd.tell() != offset:
            fd.seek(offset, 0)
        assert fd.tell() == offset
        r = fd.read(data_len)
        assert len(r) == data_len
        if needs_decoding:
            return r.decode()
        return r



"""
repository modifications (installing, removing, replacing)
"""

__all__ = (
    "Failure", "base", "install", "uninstall", "replace",
    "operations", "operations_proxy"
)

from functools import partial

from snakeoil import klass
from snakeoil.currying import post_curry
from snakeoil.dependant_methods import ForcedDepends

from .. import operations as operations_mod
from ..exceptions import PkgcoreException
from ..log import logger
from ..package.mutated import MutatedPkg
from ..restrictions import packages
from ..sync import base as _sync_base
from . import observer as observer_mod
from . import regen


class fake_lock:
    def __init__(self):
        pass

    acquire_write_lock = acquire_read_lock = __init__
    release_read_lock = release_write_lock = __init__


class Failure(PkgcoreException):
    pass


class base(metaclass=ForcedDepends):

    stage_depends = {}

    def __init__(self, repo, observer):
        self.repo = repo
        self.underway = False
        self.observer = observer
        try:
            self.lock = getattr(repo, "lock")
        except AttributeError:
            raise
        if self.lock is None:
            self.lock = fake_lock()

    def start(self, *args):
        self.underway = True
        self.lock.acquire_write_lock()
        return True

    def finalize_data(self):
        raise NotImplementedError(self, 'finalize_data')

    def finish(self):
        self.lock.release_write_lock()
        self.underway = False
        return True


class install(base):

    stage_depends = {
        'finish': '_notify_repo_add',
        '_notify_repo_add': 'finalize_data',
        'finalize_data': 'add_data',
        'add_data': 'start'
    }

    description = "install"

    def __init__(self, repo, pkg, observer):
        base.__init__(self, repo, observer)
        self.new_pkg = pkg

    def _notify_repo_add(self):
        self.repo.notify_add_package(self.new_pkg)
        return True

    def add_data(self):
        raise NotImplementedError(self, 'add_data')

    def _update_pkg_contents(self, contents):
        self.new_pkg = MutatedPkg(self.new_pkg, {"contents": contents})


class uninstall(base):

    stage_depends = {
        'finish': '_notify_repo_remove',
        '_notify_repo_remove': 'finalize_data',
        'finalize_data': 'remove_data',
        'remove_data': 'start'
    }

    description = "uninstall"

    def __init__(self, repo, pkg, observer):
        base.__init__(self, repo, observer)
        self.old_pkg = pkg

    def _notify_repo_remove(self):
        self.repo.notify_remove_package(self.old_pkg)
        return True

    def remove_data(self):
        raise NotImplementedError(self, 'remove_data')


class replace(install, uninstall):

    stage_depends = {
        'finish': '_notify_repo_add',
        '_notify_repo_add': 'finalize_data',
        'finalize_data': ('add_data', '_notify_repo_remove'),
        '_notify_repo_remove': 'remove_data',
        'remove_data': 'start',
        'add_data': 'start'
    }

    description = "replace"

    def __init__(self, repo, oldpkg, newpkg, observer):
        # yes there is duplicate initialization here.
        uninstall.__init__(self, repo, oldpkg, observer)
        install.__init__(self, repo, newpkg, observer)


class sync_operations(operations_mod.base):

    def __init__(self, repository, disable_overrides=(), enable_overrides=()):
        self.repo = repository
        super().__init__(disable_overrides, enable_overrides)

    @operations_mod.is_standalone
    def _cmd_api_sync(self, observer=None, **kwargs):
        # often enough, the syncer is a lazy_ref
        syncer = self._get_syncer()
        self.repo._pre_sync()
        ret = syncer.sync(**kwargs)
        self.repo._post_sync()
        return ret

    def _get_syncer(self, lazy=False):
        syncer = getattr(self.repo, '_syncer', klass.sentinel)
        if syncer is klass.sentinel:
            # raw repo's vs non-raw; drive down to the raw repo.
            # see pkgcore.ebuild.repository for an example
            syncer = getattr(self.repo, 'config', None)
            syncer = getattr(syncer, '_syncer', None)

        if not lazy and not isinstance(syncer, _sync_base.Syncer):
            syncer = syncer.instantiate()
        return syncer

    def _cmd_check_support_sync(self):
        syncer = self._get_syncer(lazy=True)
        if syncer is not None:
            return not self._get_syncer().disabled
        return False


class operations(sync_operations):

    def _disabled_if_frozen(self, command):
        if self.repo.frozen:
            logger.debug(
                "disabling repo(%r) command(%r) due to repo being frozen",
                self.repo, command)
        return not self.repo.frozen

    def _get_observer(self, observer=None):
        if observer is None:
            observer = observer_mod.repo_observer(observer_mod.null_output())
        return observer

    def _cmd_api_install(self, pkg, observer=None):
        return self._cmd_implementation_install(
            pkg, self._get_observer(observer))

    def _cmd_api_uninstall(self, pkg, observer=None):
        return self._cmd_implementation_uninstall(
            pkg, self._get_observer(observer))

    def _cmd_api_replace(self, oldpkg, newpkg, observer=None):
        return self._cmd_implementation_replace(
            oldpkg, newpkg, self._get_observer(observer))

    def _cmd_api_install_or_replace(self, newpkg, observer=None):
        return self._cmd_implementation_install_or_replace(
            newpkg, self._get_observer(observer))

    def _cmd_implementation_install_or_replace(self, newpkg, observer=None):
        match = self.repo.match(newpkg.versioned_atom)
        if not match:
            return self.install(newpkg, observer=observer)
        assert len(match) == 1
        return self.replace(match[0], newpkg, observer=observer)

    for x in ("install", "uninstall", "replace", "install_or_replace"):
        locals()["_cmd_check_support_%s" % x] = post_curry(
            _disabled_if_frozen, x)

    del x

    def _cmd_api_configure(self, pkg, observer=None):
        return self._cmd_implementation_configure(
            self.repo, pkg, self._get_observer(observer))

    def _cmd_implementation_clean_cache(self, pkgs=None):
        """Clean stale and invalid cache entries up."""
        caches = [x for x in self._get_caches() if not x.readonly]
        if not caches:
            return
        if pkgs is None:
            pkgs = frozenset(pkg.cpvstr for pkg in self.repo)
        for cache in caches:
            cache_pkgs = frozenset(cache)
            for p in cache_pkgs - pkgs:
                del cache[p]

    @operations_mod.is_standalone
    def _cmd_api_regen_cache(self, observer=None, threads=1, **kwargs):
        cache = getattr(self.repo, 'cache', None)
        if not cache and not kwargs.get('force', False):
            return
        sync_rate = getattr(cache, 'sync_rate', None)
        try:
            if sync_rate is not None:
                cache.set_sync_rate(1000000)
            errors = 0

            # Force usage of unfiltered repo to include pkgs with metadata issues.
            # Matches are collapsed directly to a list to avoid threading issues such
            # as EBADF since the repo iterator isn't thread-safe.
            pkgs = list(self.repo.itermatch(packages.AlwaysTrue, pkg_filter=None))

            observer = self._get_observer(observer)
            for pkg, e in regen.regen_repository(
                    self.repo, pkgs, observer=observer, threads=threads, **kwargs):
                observer.error(f'caught exception {e} while processing {pkg.cpvstr}')
                errors += 1

            # report pkgs with bad metadata -- relies on iterating over the
            # unfiltered repo to populate the masked repo
            pkgs = frozenset(pkg.cpvstr for pkg in self.repo)
            for pkg in sorted(self.repo._bad_masked):
                observer.error(f'{pkg.cpvstr}: {pkg.data.msg(verbosity=observer.verbosity)}')
                errors += 1

            # remove old/invalid cache entries
            self._cmd_implementation_clean_cache(pkgs)

            return errors
        finally:
            if sync_rate is not None:
                cache.set_sync_rate(sync_rate)
            self.repo.operations.run_if_supported("flush_cache")

    def _get_caches(self):
        caches = getattr(self.repo, 'cache', ())
        if not hasattr(caches, 'commit'):
            return caches
        return [caches]

    @operations_mod.is_standalone
    def _cmd_api_flush_cache(self, observer=None):
        for cache in self._get_caches():
            cache.commit(force=True)

    def _cmd_api_manifest(self, domain, restriction, observer=None, **kwargs):
        observer = self._get_observer(observer)
        return self._cmd_implementation_manifest(domain, restriction, observer, **kwargs)


class operations_proxy(operations):

    # cache this; this is to prevent the target operations mutating resulting
    # in our proxy setup not matching the target.
    @klass.cached_property
    def raw_operations(self):
        return self.repo.raw_repo.operations

    @klass.cached_property
    def enabled_operations(self):
        s = set(self.raw_operations.enabled_operations)
        return frozenset(self._apply_overrides(s))

    def _setup_api(self):
        for op in self.raw_operations.enabled_operations:
            setattr(self, op, partial(self._proxy_op, op))

    def _proxy_op(self, op_name, *args, **kwargs):
        return getattr(self.raw_operations, op_name)(*args, **kwargs)



import os
from functools import partial

from snakeoil import osutils
from snakeoil.cli import arghparse

from ..restrictions.boolean import AndRestriction
from ..util import commandline, packages
from . import atom, conditionals
from .eapi import get_eapi


def str_pkg(pkg):
    pkg = packages.get_raw_pkg(pkg)
    # special casing; old style virtuals come through as the original pkg.
    if pkg.package_is_real:
        return pkg.cpvstr
    if hasattr(pkg, "actual_pkg"):
        return pkg.actual_pkg.cpvstr
    # icky, but works.
    return str(pkg.rdepend).lstrip("=")


def get_atom_kls(value):
    eapi = get_eapi(value)
    if eapi is None:
        raise ValueError(f"EAPI {value} isn't known/supported")
    return eapi.atom_kls

def default_portageq_args(parser):
    parser.add_argument("--eapi", dest='atom_kls', type=get_atom_kls,
        default=atom.atom,
        help="limit all operations to just what the given EAPI supports.")
    parser.add_argument("--use", default=None,
        help="override the use flags used for transititive USE deps- "
       "dev-lang/python[threads=] for example")


def make_atom(value):
    return arghparse.DelayedValue(partial(_render_atom, value), 100)

def _render_atom(value, namespace, attr):
    a = namespace.atom_kls(value)
    if isinstance(a, atom.transitive_use_atom):
        a.restrictions
        # XXX bit of a hack.
        a = conditionals.DepSet(a.restrictions, atom.atom, True)
        a = a.evaluate_depset(getattr(namespace, 'use', ()))
        a = AndRestriction(*a.restrictions)
    setattr(namespace, attr, a)


class BaseCommand(arghparse.ArgparseCommand):

    required_arg_count = 0
    has_optional_args = False
    arg_spec = ()

    def bind_to_parser(self, parser, compat=False):
        arghparse.ArgparseCommand.bind_to_parser(self, parser)
        default_portageq_args(parser)

        if self.requires_root:
            if compat:
                kwds = {}
                if self._compat_root_default:
                    kwds["nargs"] = "?"
                    kwds["default"] = self._compat_root_default
                parser.add_argument(
                    dest="domain", metavar="root",
                    action=commandline.DomainFromPath,
                    help="the domain that lives at root will be used", **kwds)
            else:
                mux = parser.add_mutually_exclusive_group()
                commandline._mk_domain(mux)
                mux.add_argument(
                    '--domain-at-root', action=commandline.DomainFromPath,
                    dest="domain", help="specify the domain to use via its root path")

        for token in self.arg_spec:
            kwds = {}
            if token[-1] in '+?*':
                kwds["nargs"] = token[-1]
                token = token[:-1]
            if token == 'atom':
                parser.add_argument(
                    'atom', help="atom to inspect",
                    type=make_atom, **kwds)
            else:
                parser.add_argument(
                    token, help=f"{token} to inspect", **kwds)

    @classmethod
    def make_command(cls, arg_spec='', requires_root=True, bind=None,
                     root_default=None, name=None, **kwds):
        kwds = dict(
            arg_spec=tuple(arg_spec.split()), requires_root=requires_root,
            _compat_root_default=root_default, **kwds)

        def internal_function(functor, name=name):
            class mycommand(BaseCommand):
                function = __call__ = staticmethod(functor)
                __doc__ = getattr(functor, '__doc__', None)
                locals().update(kwds)

            if name is None:
                name = functor.__name__
            mycommand.__name__ = name

            if bind is not None:
                bind.append(mycommand)
            return mycommand

        return internal_function


common_commands = []
query_commands = []
portageq_commands = []

@BaseCommand.make_command("variable+", bind=query_commands)
def env_var(options, out, err):
    """
    return configuration defined variables.
    """
    for x in options.variable:
        val = options.domain.settings.get(x, '')
        if not isinstance(val, str):
            val = ' '.join(val)
        out.write(str(val))
    return 0

@BaseCommand.make_command("variable+", bind=portageq_commands, name='envvar',
    root_default='/')
def portageq_envvar(options, out, err):
    """
    return configuration defined variables.  Use envvar2 instead, this will be removed.
    """
    return env_var.function(options, out, err)

@BaseCommand.make_command("variable+", bind=portageq_commands, name='envvar2')
def portageq_envvar2(options, out, err):
    """
    return configuration defined variables.
    """
    return env_var.function(options, out, err)


@BaseCommand.make_command("atom", bind=common_commands)
def has_version(options, out, err):
    """
    Return 0 if an atom is merged, 1 if not.
    """
    if options.atom in options.domain.all_installed_repos:
        return 0
    return 1


def _best_version(domain, restrict):
    try:
        p = max(domain.all_installed_repos.itermatch(restrict))
    except ValueError:
        # empty sequence.
        return ''
    return str_pkg(p)

@BaseCommand.make_command("atom+", bind=common_commands)
def mass_best_version(options, out, err):
    """
    multiple best_version calls.
    """
    for x in options.atom:
        out.write("%s:%s" %
            (x, _best_version(options.domain, x).rstrip()))
    return 0

@BaseCommand.make_command("atom", bind=common_commands)
def best_version(options, out, err):
    """
    Return the maximum visible version for a given atom.
    """
    out.write(_best_version(options.domain, options.atom))
    return 0


@BaseCommand.make_command("atom", bind=portageq_commands)
def match(options, out, err):
    """shorthand for `pquery --installed`"""
    i = options.domain.all_installed_repos.itermatch(options.atom, sorter=sorted)
    for pkg in i:
        out.write(str_pkg(pkg))
    return 0


@BaseCommand.make_command(bind=common_commands, root_default='/')
def get_repos(options, out, err):
    l = []
    for repo in options.domain.ebuild_repos_raw:
        repo_id = getattr(repo, 'repo_id', getattr(repo, 'location', None))
        l.append(repo_id)
    for x in sorted(set(l)):
        out.write(x)
    return 0


def find_profile_paths_by_repo_id(config, repo_id, fullpath=False):
    repo = config.repo.get(repo_id, None)
    if repo is not None and getattr(repo, 'location', None) is not None:
        profiles = repo.config.profiles.arch_profiles
        for arch in profiles.keys():
            for path, stability in profiles[arch]:
                if fullpath:
                    path = os.path.join(repo.location, 'profiles', path)
                yield path


@BaseCommand.make_command("repo_id", bind=query_commands)
def get_profiles(options, out, err):
    if options.repo_id == 'all':
        profiles = (
            profile for repo in options.domain.ebuild_repos_raw
            for profile in find_profile_paths_by_repo_id(
                options.config, repo.repo_id, fullpath=True))
    else:
        profiles = find_profile_paths_by_repo_id(options.config, options.repo_id)
    for x in sorted(set(profiles)):
        out.write(x)
    return 0


@BaseCommand.make_command("repo_id", bind=portageq_commands)
def get_repo_path(options, out, err):
    repo = options.config.repo.get(options.repo_id, None)
    if repo is not None and getattr(repo, 'location', None) is not None:
        out.write(repo.location)
        return 0
    return 1

get_repo_path = BaseCommand.make_command(
    "repo_id", bind=query_commands, name='get_repo_path')(get_repo_path.function)


@BaseCommand.make_command("repo_id", bind=portageq_commands)
def get_repo_news_path(options, out, err):
    repo = options.config.repo.get(options.repo_id, None)
    if repo is not None and getattr(repo, 'location', None) is not None:
        out.write(osutils.normpath(osutils.pjoin(repo.location, 'metadata', 'news')))
        return 0
    return 1


@BaseCommand.make_command("root? repo_id", bind=portageq_commands,
    requires_root=False, name='get_repo_news_path')
def portageq_get_repo_news_path(options, out, err):
    return get_repo_news_path.function(options, out, err)

def bind_parser(parser, compat=False, name='portageq'):
    subparsers = parser.add_subparsers(description=f"{name} commands")
    l = common_commands[:]
    if compat:
        l += portageq_commands
    else:
        l += query_commands

    for command in sorted(l, key=lambda x:x.__name__):
        subparser = subparsers.add_parser(
            command.__name__, help=command.__doc__, description=command.__doc__)
        command().bind_to_parser(subparser, compat=compat)



import difflib

from pkgcore.ebuild.atom import atom
from pkgcore.ebuild.formatter import BasicFormatter, PkgcoreFormatter, PortageFormatter
from pkgcore.test.misc import FakePkg, FakeRepo
from snakeoil.test import TestCase
from snakeoil.test.argparse_helpers import Bold, Color, FakeStreamFormatter, Reset


# These two are probably unnecessary with ferringb's changes to
# PkgcoreFormatter, but as he's the one that uses it there's no point fixing
# what ain't broke. --charlie
class FakeMutatedPkg(FakePkg):
    def __str__(self):
        # Yes this should be less hackish (and hardcoded values suck),
        # but we can't really subclass MutatedPkg so this will have to do
        return f"MutatedPkg(built ebuild: {self.cpvstr}, overrides=('depend', 'rdepend'))"

class FakeEbuildSrc(FakePkg):
    def __str__(self):
        # Yes this should be less hackish (and hardcoded values suck)
        # but we can't really subclass ebuild_src so this will have to do
        return f"config wrapped(use): ebuild src: {self.cpvstr}"


class FakeOp:
    def __init__(self, package, oldpackage=None, desc='add'):
        self.pkg = package
        if oldpackage:
            self.old_pkg = oldpackage
            self.desc = 'replace'
        else:
            self.desc = desc

class BaseFormatterTest:
    prefix = ()
    suffix = ('\n',)
    def setUp(self):
        self.fakeout = FakeStreamFormatter()
        self.fakeerr = FakeStreamFormatter()
        self.formatter = self.newFormatter()

    @property
    def verify_formatterClass(self):
        class state_verifying_class(self.formatterClass):
            def format(internal_self, *args, **kwds):
                autoline = self.fakeout.autoline
                try:
                    ret = self.formatterClass.format(internal_self, *args, **kwds)
                except Exception as e:
                    self.assertEqual(
                        autoline, self.fakeout.autoline,
                        msg="exception thrown {e}, autoline was {autoline}, now is {self.fakeout.autoline}")
                    raise
                self.assertEqual(
                    autoline, self.fakeout.autoline,
                    msg="autoline was {autoline}, now is {self.fakeout.autoline}")
                return ret
        return state_verifying_class

    def newFormatter(self, **kwargs):
        disable_method_checks = kwargs.pop("disable_method_checks", False)
        kwargs.setdefault("out", self.fakeout)
        kwargs.setdefault("err", self.fakeerr)
        if not disable_method_checks:
            kls = self.verify_formatterClass
        else:
            kls = self.formatterClass
        return kls(**kwargs)

    def assertOut(self, *args, **kwargs):

        stringlist = []
        objectlist = []

        args = list(args)

        prefix = kwargs.setdefault("prefix", self.prefix)
        if isinstance(prefix, tuple):
            args = list(prefix) + args
        elif isinstance(prefix, list):
            args = prefix + args
        else:
            args.insert(0, prefix)

        suffix = kwargs.setdefault("suffix", self.suffix)
        if isinstance(suffix, tuple):
            args = args + list(suffix)
        elif isinstance(suffix, list):
            args = args + suffix
        else:
            args.append(suffix)

        for arg in args:
            if isinstance(arg, str):
                stringlist.append(arg.encode('ascii'))
            elif isinstance(arg, bytes):
                stringlist.append(arg)
            else:
                objectlist.append(b''.join(stringlist))
                stringlist = []
                objectlist.append(arg)
        objectlist.append(b''.join(stringlist))

        # Hack because a list with an empty string in is True
        if objectlist == [b'']:
            objectlist = []

        self.assertEqual(self.fakeout.stream, objectlist, '\n' + '\n'.join(
                difflib.unified_diff(
                    list(repr(s) for s in objectlist),
                    list(repr(s) for s in self.fakeout.stream),
                    'expected', 'actual', lineterm='')))
        self.fakeout.resetstream()

    def test_end(self):
        # Sub-classes should override this if they print something in end()
        self.formatter.format(FakeOp(FakeMutatedPkg('dev-util/diffball-1.1')))
        self.fakeout.resetstream()
        self.formatter.end()
        self.assertOut(suffix=())


class TestBasicFormatter(BaseFormatterTest, TestCase):

    formatterClass = BasicFormatter

    def test_install(self):
        # Make sure we ignore versions...
        self.formatter.format(FakeOp(FakeMutatedPkg('dev-util/diffball-1.1')))
        self.assertOut('dev-util/diffball')

    def test_reinstall(self):
        self.formatter.format(
            FakeOp(FakeEbuildSrc('app-arch/bzip2-1.0.4'),
            FakeMutatedPkg('app-arch/bzip2-1.0.4')))
        self.assertOut('app-arch/bzip2')


class TestPkgcoreFormatter(BaseFormatterTest, TestCase):

    formatterClass = PkgcoreFormatter

    def test_install(self):
        self.formatter.format(FakeOp(FakeEbuildSrc('dev-util/diffball-1.0')))
        self.assertOut("add     dev-util/diffball-1.0")

    def test_install_repo(self):
        self.formatter.format(FakeOp(FakeEbuildSrc('dev-util/diffball-1.0',
            repo=FakeRepo(repo_id='gentoo', location='/var/gentoo/repos/gentoo'))))
        self.assertOut("add     dev-util/diffball-1.0::gentoo")

    def test_reinstall(self):
        self.formatter.format(
            FakeOp(FakeEbuildSrc('dev-util/diffball-1.2'),
                FakeMutatedPkg('dev-util/diffball-1.1')))
        self.assertOut(
            "replace dev-util/diffball-1.1, "
            "dev-util/diffball-1.2")

    def test_reinstall_repo(self):
        self.formatter.format(
            FakeOp(FakeEbuildSrc('dev-util/diffball-1.2',
                   repo=FakeRepo(repo_id='gentoo', location='/var/gentoo/repos/gentoo')),
                FakeMutatedPkg('dev-util/diffball-1.1')))
        self.assertOut(
            "replace dev-util/diffball-1.1, "
            "dev-util/diffball-1.2::gentoo")


class CountingFormatterTest(BaseFormatterTest):

    endprefix = ""
    endsuffix = "\n"

    def newFormatter(self, **kwargs):
        kwargs.setdefault('verbosity', 1)
        return BaseFormatterTest.newFormatter(self, **kwargs)

    def assertEnd(self, *args, **kwargs):
        kwargs.setdefault('prefix', self.endprefix)
        kwargs.setdefault('suffix', self.endsuffix)
        BaseFormatterTest.assertOut(self, *args, **kwargs)

    def test_end(self):
        self.formatter.end()
        self.assertEnd('\nTotal: 0 packages')

    def test_end_new(self):
        self.formatter.format(FakeOp(FakeEbuildSrc('app-arch/bzip2-1.0.3-r6')))
        self.fakeout.resetstream()
        self.formatter.end()
        self.assertEnd('\nTotal: 1 package (1 new)')

    def test_end_new_multiple(self):
        self.formatter.format(FakeOp(FakeEbuildSrc('app-arch/bzip2-1.0.3-r6')))
        self.formatter.format(FakeOp(FakeEbuildSrc('app-arch/gzip-1.6')))
        self.fakeout.resetstream()
        self.formatter.end()
        self.assertEnd('\nTotal: 2 packages (2 new)')

    def test_end_newslot(self):
        self.formatter.format(FakeOp(FakeEbuildSrc('app-arch/bzip2-1.0.3-r6', slot='1')))
        self.fakeout.resetstream()
        self.formatter.end()
        self.assertEnd('\nTotal: 1 package (1 in new slot)')

    def test_end_newslot_multiple(self):
        self.formatter.format(FakeOp(FakeEbuildSrc('app-arch/bzip2-1.0.3-r6', slot='1')))
        self.formatter.format(FakeOp(FakeEbuildSrc('app-arch/gzip-1.6', slot='2')))
        self.fakeout.resetstream()
        self.formatter.end()
        self.assertEnd('\nTotal: 2 packages (2 in new slots)')

    def test_end_downgrade(self):
        self.formatter.format(
            FakeOp(FakeEbuildSrc('app-arch/bzip2-1.0.3-r6'),
            FakeMutatedPkg('app-arch/bzip2-1.0.4')))
        self.fakeout.resetstream()
        self.formatter.end()
        self.assertEnd('\nTotal: 1 package (1 downgrade)')

    def test_end_downgrade_multiple(self):
        self.formatter.format(
            FakeOp(FakeEbuildSrc('app-arch/bzip2-1.0.3-r6'),
            FakeMutatedPkg('app-arch/bzip2-1.0.4')))
        self.formatter.format(
            FakeOp(FakeEbuildSrc('app-arch/gzip-1.5'),
            FakeMutatedPkg('app-arch/gzip-1.6')))
        self.fakeout.resetstream()
        self.formatter.end()
        self.assertEnd('\nTotal: 2 packages (2 downgrades)')

    def test_end_upgrade(self):
        self.formatter.format(
            FakeOp(FakeEbuildSrc('app-arch/bzip2-1.0.4'),
            FakeMutatedPkg('app-arch/bzip2-1.0.3-r6')))
        self.fakeout.resetstream()
        self.formatter.end()
        self.assertEnd('\nTotal: 1 package (1 upgrade)')

    def test_end_upgrade_multiple(self):
        self.formatter.format(
            FakeOp(FakeEbuildSrc('app-arch/bzip2-1.0.4'),
            FakeMutatedPkg('app-arch/bzip2-1.0.3-r6')))
        self.formatter.format(
            FakeOp(FakeEbuildSrc('app-arch/gzip-1.6'),
            FakeMutatedPkg('app-arch/gzip-1.5')))
        self.fakeout.resetstream()
        self.formatter.end()
        self.assertEnd('\nTotal: 2 packages (2 upgrades)')

    def test_end_reinstall(self):
        self.formatter.format(
            FakeOp(FakeEbuildSrc('app-arch/bzip2-1.0.4'),
            FakeMutatedPkg('app-arch/bzip2-1.0.4')))
        self.fakeout.resetstream()
        self.formatter.end()
        self.assertEnd('\nTotal: 1 package (1 reinstall)')

    def test_end_reinstall_multiple(self):
        self.formatter.format(
            FakeOp(FakeEbuildSrc('app-arch/bzip2-1.0.4'),
            FakeMutatedPkg('app-arch/bzip2-1.0.4')))
        self.formatter.format(
            FakeOp(FakeEbuildSrc('app-arch/gzip-1.6'),
            FakeMutatedPkg('app-arch/gzip-1.6')))
        self.fakeout.resetstream()
        self.formatter.end()
        self.assertEnd('\nTotal: 2 packages (2 reinstalls)')

    def test_end_all_ops_order(self):
        # new
        self.formatter.format(
            FakeOp(FakeEbuildSrc('app-arch/pkga-1.0.3-r6')))
        # new slot
        self.formatter.format(
            FakeOp(FakeEbuildSrc('app-arch/pkgb-1.0.3-r6', slot='1')))
        # downgrade
        self.formatter.format(
            FakeOp(FakeEbuildSrc('app-arch/pkgc-1.0.3-r6'),
            FakeMutatedPkg('app-arch/pkgc-1.0.4')))
        # upgrade
        self.formatter.format(
            FakeOp(FakeEbuildSrc('app-arch/pkgd-1.0.4'),
            FakeMutatedPkg('app-arch/pkgd-1.0.3-r6')))
        # reinstall
        self.formatter.format(
            FakeOp(FakeEbuildSrc('app-arch/pkge-1.0.4'),
            FakeMutatedPkg('app-arch/pkge-1.0.4')))

        self.fakeout.resetstream()
        self.formatter.end()
        self.assertEnd(
            '\nTotal: 5 packages (1 new, 1 upgrade, 1 downgrade, 1 in new slot, 1 reinstall)')


class TestPortageFormatter(BaseFormatterTest, TestCase):

    formatterClass = PortageFormatter

    def setUp(self):
        pkg = FakeMutatedPkg('app-arch/bzip2-1.0.1-r1', slot='0')
        masked_atom = atom('>=app-arch/bzip2-2.0')
        self.domain_settings = {"ACCEPT_KEYWORDS": ("amd64",)}
        self.repo1 = FakeRepo(
            repo_id='gentoo', location='/var/gentoo/repos/gentoo',
            masks=(masked_atom,), domain_settings=self.domain_settings)
        self.repo2 = FakeRepo(
            repo_id='repo2', location='/var/gentoo/repos/repo2',
            domain_settings=self.domain_settings)
        self.vdb = FakeRepo(repo_id='vdb', pkgs=[pkg])
        BaseFormatterTest.setUp(self)

    def newFormatter(self, **kwargs):
        kwargs.setdefault('quiet_repo_display', False)
        kwargs.setdefault('installed_repos', self.vdb)
        return BaseFormatterTest.newFormatter(self, **kwargs)

    def repo_id(self, repo):
        if getattr(self.formatter, 'verbosity', 0):
            return '::' + repo.repo_id
        return ''

    def test_new(self):
        self.formatter.format(
            FakeOp(FakeEbuildSrc('app-arch/bzip2-1.0.4')))
        self.assertOut('[', Color('fg', 'green'), 'ebuild', Reset(),
            '  ', Color('fg', 'green'), Bold(), 'N', Reset(), '     ] ',
            Color('fg', 'green'), 'app-arch/bzip2-1.0.4', Reset())

    def test_remove(self):
        self.formatter.format(
            FakeOp(FakeEbuildSrc('app-arch/bzip2-1.0.4'), desc='remove'))
        self.assertOut('[', Color('fg', 'red'), 'uninstall', Reset(),
            '        ] ', Color('fg', 'red'), 'app-arch/bzip2-1.0.4', Reset())

    def test_upgrade(self):
        self.formatter.format(
            FakeOp(FakeEbuildSrc('app-arch/bzip2-1.0.4'),
            FakeMutatedPkg('app-arch/bzip2-1.0.3-r6')))
        self.assertOut('[', Color('fg', 'green'), 'ebuild', Reset(),
            '     ', Color('fg', 'cyan'), Bold(), 'U', Reset(), '  ] ',
            Color('fg', 'green'), 'app-arch/bzip2-1.0.4', Reset(), ' ',
            Color('fg', 'blue'), Bold(), '[1.0.3-r6]', Reset())

    def test_downgrade(self):
        self.formatter.format(
            FakeOp(FakeEbuildSrc('app-arch/bzip2-1.0.3-r6'),
            FakeMutatedPkg('app-arch/bzip2-1.0.4')))
        self.assertOut('[', Color('fg', 'green'), 'ebuild', Reset(),
            '     ', Color('fg', 'cyan'), Bold(), 'U', Reset(),
            Color('fg', 'blue'), Bold(), 'D' , Reset(), ' ] ',
            Color('fg', 'green'), 'app-arch/bzip2-1.0.3-r6', Reset(), ' ',
            Color('fg', 'blue'), Bold(), '[1.0.4]', Reset())

    def test_reinstall(self):
        self.formatter.format(
            FakeOp(FakeEbuildSrc('app-arch/bzip2-1.0.3-r6', repo=self.repo1),
            FakeMutatedPkg('app-arch/bzip2-1.0.3-r6', repo=self.repo1)))
        self.assertOut('[', Color('fg', 'green'), 'ebuild', Reset(),
            '   ', Color('fg', 'yellow'), Bold(), 'R', Reset(), '    ] ',
            Color('fg', 'green'), f'app-arch/bzip2-1.0.3-r6{self.repo_id(self.repo1)}', Reset())

    def test_reinstall_from_new_repo(self):
        self.formatter.format(
            FakeOp(FakeEbuildSrc('app-arch/bzip2-1.0.3-r6', repo=self.repo1),
            FakeMutatedPkg('app-arch/bzip2-1.0.3-r6', repo=self.repo2)))
        self.assertOut('[', Color('fg', 'green'), 'ebuild', Reset(),
            '   ', Color('fg', 'yellow'), Bold(), 'R', Reset(), '    ] ',
            Color('fg', 'green'), f'app-arch/bzip2-1.0.3-r6{self.repo_id(self.repo1)}', Reset(),
            ' ', Color('fg', 'blue'), Bold(), f'[1.0.3-r6{self.repo_id(self.repo2)}]', Reset())

    def test_new_use(self):
        self.formatter.format(
            FakeOp(FakeEbuildSrc('app-arch/bzip2-1.0.3-r6', iuse=['static'], use=['static'])))
        self.assertOut('[', Color('fg', 'green'), 'ebuild', Reset(),
            '  ', Color('fg', 'green'), Bold(), 'N', Reset(), '     ] ',
            Color('fg', 'green'), 'app-arch/bzip2-1.0.3-r6', Reset(),
            ' USE="', Color('fg', 'red'), Bold(), 'static', Reset(), '"')

    def test_new_nouse(self):
        self.formatter.format(
            FakeOp(FakeEbuildSrc('app-arch/bzip2-1.0.3-r6', iuse=['static'])))
        self.assertOut('[', Color('fg', 'green'), 'ebuild', Reset(),
            '  ', Color('fg', 'green'), Bold(), 'N', Reset(), '     ] ',
            Color('fg', 'green'), 'app-arch/bzip2-1.0.3-r6', Reset(),
            ' USE="', Color('fg', 'blue'), Bold(), '-static', Reset(), '"')

    def test_nouse(self):
        self.formatter.format(
            FakeOp(FakeEbuildSrc('app-arch/bzip2-1.0.3-r6', iuse=['static']),
            FakeMutatedPkg('app-arch/bzip2-1.0.3-r6')))
        self.assertOut('[', Color('fg', 'green'), 'ebuild', Reset(),
            '   ', Color('fg', 'yellow'), Bold(), 'R', Reset(), '    ] ',
            Color('fg', 'green'), 'app-arch/bzip2-1.0.3-r6', Reset(),
            ' USE="', Color('fg', 'yellow'), Bold(), '-static', Reset(), '%"')

    def test_iuse_defaults(self):
        self.formatter.format(
            FakeOp(FakeEbuildSrc('app-arch/bzip2-1.0.3-r6', eapi='1', iuse=['+static', '-junk'], use=['static']),
            FakeMutatedPkg('app-arch/bzip2-1.0.3-r6')))
        self.assertOut('[', Color('fg', 'green'), 'ebuild', Reset(),
            '   ', Color('fg', 'yellow'), Bold(), 'R', Reset(), '    ] ',
            Color('fg', 'green'), 'app-arch/bzip2-1.0.3-r6', Reset(),
            ' USE="', Color('fg', 'yellow'), Bold(), 'static', Reset(), "%* ",
            Color('fg', 'yellow'), Bold(), '-junk', Reset(), '%"')

    def test_use_enabled(self):
        self.formatter.format(FakeOp(
            FakeEbuildSrc('app-arch/bzip2-1.0.3-r6', iuse=['static'], use=['static']),
            FakeMutatedPkg('app-arch/bzip2-1.0.3-r6', iuse=['static']),
        ))
        self.assertOut('[', Color('fg', 'green'), 'ebuild', Reset(),
            '   ', Color('fg', 'yellow'), Bold(), 'R', Reset(), '    ] ',
            Color('fg', 'green'), 'app-arch/bzip2-1.0.3-r6', Reset(),
            ' USE="', Color('fg', 'green'), Bold(), 'static', Reset(), '*"')

    def test_use_disabled(self):
        self.formatter.format(FakeOp(
            FakeEbuildSrc('app-arch/bzip2-1.0.3-r6', iuse=['static']),
            FakeMutatedPkg('app-arch/bzip2-1.0.3-r6', iuse=['static'], use=['static']),
        ))
        self.assertOut('[', Color('fg', 'green'), 'ebuild', Reset(),
            '   ', Color('fg', 'yellow'), Bold(), 'R', Reset(), '    ] ',
            Color('fg', 'green'), 'app-arch/bzip2-1.0.3-r6', Reset(),
            ' USE="', Color('fg', 'green'), Bold(), '-static', Reset(), '*"')

    def test_multiuse(self):
        self.formatter.format(
            FakeOp(FakeEbuildSrc('app-arch/bzip2-1.0.3-r6', iuse=['static', 'bootstrap'], use=['static']),
            FakeMutatedPkg('app-arch/bzip2-1.0.3-r6')))
        self.assertOut('[', Color('fg', 'green'), 'ebuild', Reset(),
            '   ', Color('fg', 'yellow'), Bold(), 'R', Reset(), '    ] ',
            Color('fg', 'green'), 'app-arch/bzip2-1.0.3-r6', Reset(),
            ' USE="', Color('fg', 'yellow'), Bold(), 'static', Reset(), '%* ',
            Color('fg', 'yellow'), Bold(), '-bootstrap', Reset(), '%"')

    def test_misc(self):
        self.formatter.format(FakeOp(FakeEbuildSrc('app-arch/bzip2-1.0.3-r6', slot='1')))
        self.assertOut('[', Color('fg', 'green'), 'ebuild', Reset(),
            '  ', Color('fg', 'green'), Bold(), 'N', Reset(), Color('fg', 'green'), Bold(),
            'S', Reset(), '    ] ',
            Color('fg', 'green'), 'app-arch/bzip2-1.0.3-r6', Reset(),
            ' ', Color('fg', 'blue'), Bold(), '[1.0.1-r1]', Reset())

    def test_fetch_restrict_no_fetchables(self):
        self.formatter.format(FakeOp(FakeEbuildSrc('app-arch/bzip2-1.0.3-r6', restrict='fetch')))
        self.assertOut('[', Color('fg', 'green'), 'ebuild', Reset(),
            '  ', Color('fg', 'green'), Bold(), 'N', Reset(),
            ' ', Color('fg', 'green'), Bold(), 'f', Reset(), '   ] ',
            Color('fg', 'green'), 'app-arch/bzip2-1.0.3-r6', Reset())

    # TODO
    def test_fetch_restrict_missing_fetchables(self):
        pass

    # TODO
    def test_fetch_restrict_prefetched_fetchables(self):
        pass

    def test_added_iuse_disabled(self):
        self.formatter.format(FakeOp(
            FakeEbuildSrc('app-arch/bzip2-1.0.3-r6', iuse=['bootstrap']),
            FakeMutatedPkg('app-arch/bzip2-1.0.3-r6'),
        ))
        self.assertOut('[', Color('fg', 'green'), 'ebuild', Reset(),
            '   ', Color('fg', 'yellow'), Bold(), 'R', Reset(), '    ] ',
            Color('fg', 'green'), 'app-arch/bzip2-1.0.3-r6', Reset(),
            ' USE="', Color('fg', 'yellow'), Bold(), '-bootstrap', Reset(), '%"')

    def test_added_iuse_enabled(self):
        self.formatter.format(FakeOp(
            FakeEbuildSrc('app-arch/bzip2-1.0.3-r6', iuse=['static'], use=['static']),
            FakeMutatedPkg('app-arch/bzip2-1.0.3-r6'),
        ))
        self.assertOut('[', Color('fg', 'green'), 'ebuild', Reset(),
            '   ', Color('fg', 'yellow'), Bold(), 'R', Reset(), '    ] ',
            Color('fg', 'green'), 'app-arch/bzip2-1.0.3-r6', Reset(),
            ' USE="', Color('fg', 'yellow'), Bold(), 'static', Reset(), '%*"')

    def test_dropped_iuse_disabled(self):
        self.formatter.format(FakeOp(
            FakeEbuildSrc('app-arch/bzip2-1.0.3-r6'),
            FakeMutatedPkg('app-arch/bzip2-1.0.3-r6', iuse=['bootstrap']),
        ))
        self.assertOut('[', Color('fg', 'green'), 'ebuild', Reset(),
            '   ', Color('fg', 'yellow'), Bold(), 'R', Reset(), '    ] ',
            Color('fg', 'green'), 'app-arch/bzip2-1.0.3-r6', Reset())

    def test_dropped_iuse_enabled(self):
        self.formatter.format(FakeOp(
            FakeEbuildSrc('app-arch/bzip2-1.0.3-r6'),
            FakeMutatedPkg('app-arch/bzip2-1.0.3-r6', iuse=['static'], use=['static']),
        ))
        self.assertOut('[', Color('fg', 'green'), 'ebuild', Reset(),
            '   ', Color('fg', 'yellow'), Bold(), 'R', Reset(), '    ] ',
            Color('fg', 'green'), 'app-arch/bzip2-1.0.3-r6', Reset())

    def test_use_expand(self):
        self.formatter = self.newFormatter(use_expand=set(["foo", "bar"]))
        self.formatter.format(
            FakeOp(FakeEbuildSrc('app-arch/bzip2-1.0.3-r6',
                iuse=['foo_static', 'foo_bootstrap', 'bar_baz'],
                use=['foo_static', 'bar_baz']),
            FakeMutatedPkg('app-arch/bzip2-1.0.3-r6')))
        self.assertOut('[', Color('fg', 'green'), 'ebuild', Reset(),
            '   ', Color('fg', 'yellow'), Bold(), 'R', Reset(), '    ] ',
            Color('fg', 'green'), 'app-arch/bzip2-1.0.3-r6', Reset(),
            ' BAR="', Color('fg', 'yellow'), Bold(), 'baz', Reset(), '%*"',
            ' FOO="', Color('fg', 'yellow'), Bold(), 'static', Reset(), '%* ',
            Color('fg', 'yellow'), Bold(), '-bootstrap', Reset(), '%"')

    def test_disabled_use(self):
        self.formatter.pkg_get_use = lambda pkg: (set(), set(), set(['static']))

        self.formatter.format(
            FakeOp(FakeEbuildSrc('app-arch/bzip2-1.0.3-r6', iuse=['static', 'bootstrap'])))
        self.assertOut('[', Color('fg', 'green'), 'ebuild', Reset(),
            '  ', Color('fg', 'green'), Bold(), 'N', Reset(), '     ] ',
            Color('fg', 'green'), 'app-arch/bzip2-1.0.3-r6', Reset(),
            ' USE="',
            Color('fg', 'blue'), Bold(), '-bootstrap', Reset(), ' ',
            '(', Color('fg', 'blue'), Bold(), '-static', Reset(), ')"')

    def test_forced_use(self):
        self.formatter.pkg_get_use = lambda pkg: (set(['static']), set(), set())

        # new pkg: static use flag forced on
        self.formatter.format(
            FakeOp(FakeEbuildSrc('app-arch/bzip2-1.0.3-r6', iuse=['static'], use=['static'])))
        self.assertOut('[', Color('fg', 'green'), 'ebuild', Reset(),
            '  ', Color('fg', 'green'), Bold(), 'N', Reset(), '     ] ',
            Color('fg', 'green'), 'app-arch/bzip2-1.0.3-r6', Reset(),
            ' USE="(', Color('fg', 'red'), Bold(), 'static', Reset(), ')"')

        # rebuilt pkg: toggled static use flag forced on
        self.formatter.format(
            FakeOp(FakeEbuildSrc('app-arch/bzip2-1.0.3-r6', iuse=['static'], use=['static']),
            FakeMutatedPkg('app-arch/bzip2-1.0.3-r6', iuse=['static'])))
        self.assertOut('[', Color('fg', 'green'), 'ebuild', Reset(),
            '   ', Color('fg', 'yellow'), Bold(), 'R', Reset(), '    ] ',
            Color('fg', 'green'), 'app-arch/bzip2-1.0.3-r6', Reset(),
            ' USE="(', Color('fg', 'green'), Bold(), 'static', Reset(), '*)"')

        # rebuilt pkg: new static use flag forced on
        self.formatter.format(
            FakeOp(FakeEbuildSrc('app-arch/bzip2-1.0.3-r6', iuse=['static'], use=['static']),
            FakeMutatedPkg('app-arch/bzip2-1.0.3-r6')))
        self.assertOut('[', Color('fg', 'green'), 'ebuild', Reset(),
            '   ', Color('fg', 'yellow'), Bold(), 'R', Reset(), '    ] ',
            Color('fg', 'green'), 'app-arch/bzip2-1.0.3-r6', Reset(),
            ' USE="(', Color('fg', 'yellow'), Bold(), 'static', Reset(), '%*)"')

    def test_forced_use_expand(self):
        self.formatter = self.newFormatter(use_expand=set(["ABI_X86", "TARGETS"]))
        self.formatter.pkg_get_use = lambda pkg: (set(['targets_X86']), set(), set())

        # rebuilt pkg: new abi_x86_64 and targets_X86 USE flags,
        # with abi_x86_64 disabled and targets_X86 forced on
        self.formatter.format(
            FakeOp(FakeEbuildSrc(
                'app-arch/bzip2-1.0.3-r6',
                iuse=['abi_x86_64', 'targets_X86'],
                use=['targets_X86']),
            FakeMutatedPkg('app-arch/bzip2-1.0.3-r6')))
        self.assertOut('[', Color('fg', 'green'), 'ebuild', Reset(),
            '   ', Color('fg', 'yellow'), Bold(), 'R', Reset(), '    ] ',
            Color('fg', 'green'), 'app-arch/bzip2-1.0.3-r6', Reset(),
            ' ABI_X86="', Color('fg', 'yellow'), Bold(), '-64', Reset(), '%"',
            ' TARGETS="(', Color('fg', 'yellow'), Bold(), 'X86', Reset(), '%*)"')

    def test_worldfile_atom(self):
        self.formatter.world_list = [atom('app-arch/bzip2')]
        self.formatter.format(
        FakeOp(FakeEbuildSrc('app-arch/bzip2-1.0.3-r6')))
        self.assertOut('[', Color('fg', 'green'), Bold(), 'ebuild', Reset(),
            '  ', Color('fg', 'green'), Bold(), 'N', Reset(), '     ] ',
            Color('fg', 'green'), Bold(), 'app-arch/bzip2-1.0.3-r6', Reset())

class TestPortageVerboseFormatter(TestPortageFormatter):

    def newFormatter(self, **kwargs):
        kwargs.setdefault("verbosity", 1)
        kwargs.setdefault("unstable_arch", "~amd64")
        return TestPortageFormatter.newFormatter(self, **kwargs)

    def test_install_symbol_unkeyworded(self):
        self.formatter.format(
            FakeOp(FakeEbuildSrc('app-arch/bzip2-1.0.3-r6', repo=self.repo1, keywords=())))
        self.assertOut('[', Color('fg', 'green'), 'ebuild', Reset(),
            '  ', Color('fg', 'green'), Bold(), 'N', Reset(), '    ',
            Color('fg', 'red'), Bold(), '*', Reset(), '] ',
            Color('fg', 'green'), f'app-arch/bzip2-1.0.3-r6{self.repo_id(self.repo1)}', Reset())

    def test_install_symbol_unstable(self):
        self.formatter.format(
            FakeOp(FakeEbuildSrc('app-arch/bzip2-1.0.3-r6', repo=self.repo1, keywords=('~amd64',))))
        self.assertOut('[', Color('fg', 'green'), 'ebuild', Reset(),
            '  ', Color('fg', 'green'), Bold(), 'N', Reset(), '    ',
            Color('fg', 'yellow'), Bold(), '~', Reset(), '] ',
            Color('fg', 'green'), f'app-arch/bzip2-1.0.3-r6{self.repo_id(self.repo1)}', Reset())

    def test_install_symbol_masked(self):
        self.formatter.format(
           FakeOp(FakeEbuildSrc('app-arch/bzip2-2.1', repo=self.repo1)))
        self.assertOut('[', Color('fg', 'green'), 'ebuild', Reset(),
           '  ', Color('fg', 'green'), Bold(), 'N', Reset(), '    ',
           Color('fg', 'red'), Bold(), '#', Reset(), '] ',
           Color('fg', 'green'), f'app-arch/bzip2-2.1{self.repo_id(self.repo1)}', Reset())

    def test_repo_id(self):
        self.formatter.format(FakeOp(FakeEbuildSrc('app-arch/bzip2-1.0.3-r6', repo=self.repo1)))
        self.assertOut('[', Color('fg', 'green'), 'ebuild', Reset(),
            '  ', Color('fg', 'green'), Bold(), 'N', Reset(), '     ] ',
            Color('fg', 'green'), 'app-arch/bzip2-1.0.3-r6::gentoo', Reset())
        self.formatter.format(FakeOp(FakeEbuildSrc('app-arch/bzip2-1.0.3-r6', repo=self.repo2)))
        self.assertOut('[', Color('fg', 'green'), 'ebuild', Reset(),
            '  ', Color('fg', 'green'), Bold(), 'N', Reset(), '     ] ',
            Color('fg', 'green'), 'app-arch/bzip2-1.0.3-r6::repo2', Reset())
        self.formatter.format(
            FakeOp(FakeEbuildSrc('app-arch/bzip2-1.0.4', repo=self.repo1),
            FakeMutatedPkg('app-arch/bzip2-1.0.3-r6', repo=self.repo1)))
        self.assertOut('[', Color('fg', 'green'), 'ebuild', Reset(),
            '     ', Color('fg', 'cyan'), Bold(), 'U', Reset(), '  ] ',
            Color('fg', 'green'), 'app-arch/bzip2-1.0.4::gentoo', Reset(), ' ',
            Color('fg', 'blue'), Bold(), '[1.0.3-r6::gentoo]', Reset())
        self.formatter.format(
            FakeOp(FakeEbuildSrc('app-arch/bzip2-1.0.4', repo=self.repo2),
            FakeMutatedPkg('app-arch/bzip2-1.0.3-r6', repo=self.repo1)))
        self.assertOut('[', Color('fg', 'green'), 'ebuild', Reset(),
            '     ', Color('fg', 'cyan'), Bold(), 'U', Reset(), '  ] ',
            Color('fg', 'green'), 'app-arch/bzip2-1.0.4::repo2', Reset(), ' ',
            Color('fg', 'blue'), Bold(), '[1.0.3-r6::gentoo]', Reset())
        self.formatter.format(
            FakeOp(FakeEbuildSrc('app-arch/bzip2-1.0.4', repo=self.repo1),
            FakeMutatedPkg('app-arch/bzip2-1.0.3-r6', repo=self.repo2)))
        self.assertOut('[', Color('fg', 'green'), 'ebuild', Reset(),
            '     ', Color('fg', 'cyan'), Bold(), 'U', Reset(), '  ] ',
            Color('fg', 'green'), 'app-arch/bzip2-1.0.4::gentoo', Reset(), ' ',
            Color('fg', 'blue'), Bold(), '[1.0.3-r6::repo2]', Reset())
        self.formatter.format(
            FakeOp(FakeEbuildSrc('app-arch/bzip2-1.0.4', repo=self.repo2),
            FakeMutatedPkg('app-arch/bzip2-1.0.3-r6', repo=self.repo2)))
        self.assertOut('[', Color('fg', 'green'), 'ebuild', Reset(),
            '     ', Color('fg', 'cyan'), Bold(), 'U', Reset(), '  ] ',
            Color('fg', 'green'), 'app-arch/bzip2-1.0.4::repo2', Reset(), ' ',
            Color('fg', 'blue'), Bold(), '[1.0.3-r6::repo2]', Reset())

    def test_misc(self):
        self.formatter.format(FakeOp(FakeEbuildSrc('app-arch/bzip2-1.0.3-r6', slot='0')))
        self.assertOut('[', Color('fg', 'green'), 'ebuild', Reset(),
            '  ', Color('fg', 'green'), Bold(), 'N', Reset(), '     ] ',
            Color('fg', 'green'), 'app-arch/bzip2-1.0.3-r6', Reset())
        self.formatter.format(FakeOp(FakeEbuildSrc('app-arch/bzip2-1.0.3-r6', slot='0', subslot='0')))
        self.assertOut('[', Color('fg', 'green'), 'ebuild', Reset(),
            '  ', Color('fg', 'green'), Bold(), 'N', Reset(), '     ] ',
            Color('fg', 'green'), 'app-arch/bzip2-1.0.3-r6', Reset())
        self.formatter.format(FakeOp(FakeEbuildSrc('app-arch/bzip2-1.0.3-r6', slot='0', subslot='2')))
        self.assertOut('[', Color('fg', 'green'), 'ebuild', Reset(),
            '  ', Color('fg', 'green'), Bold(), 'N', Reset(), '     ] ',
            Color('fg', 'green'), 'app-arch/bzip2-1.0.3-r6:0/2', Reset())
        self.formatter.format(FakeOp(FakeEbuildSrc('app-arch/bzip2-1.0.3-r6', slot='foo')))
        self.assertOut('[', Color('fg', 'green'), 'ebuild', Reset(),
            '  ', Color('fg', 'green'), Bold(), 'N', Reset(), Color('fg', 'green'), Bold(),
            'S', Reset(), '    ] ',
            Color('fg', 'green'), 'app-arch/bzip2-1.0.3-r6:foo', Reset(),
            ' ', Color('fg', 'blue'), Bold(), '[1.0.1-r1:0]', Reset())
        self.formatter.format(FakeOp(FakeEbuildSrc('app-arch/bzip2-1.0.3-r6', slot='1', subslot='0')))
        self.assertOut('[', Color('fg', 'green'), 'ebuild', Reset(),
            '  ', Color('fg', 'green'), Bold(), 'N', Reset(), Color('fg', 'green'), Bold(),
            'S', Reset(), '    ] ',
            Color('fg', 'green'), 'app-arch/bzip2-1.0.3-r6:1/0', Reset(),
            ' ', Color('fg', 'blue'), Bold(), '[1.0.1-r1:0]', Reset())
        self.formatter.format(FakeOp(FakeEbuildSrc('app-arch/bzip2-1.0.3-r6', slot='2', subslot='foo')))
        self.assertOut('[', Color('fg', 'green'), 'ebuild', Reset(),
            '  ', Color('fg', 'green'), Bold(), 'N', Reset(), Color('fg', 'green'), Bold(),
            'S', Reset(), '    ] ',
            Color('fg', 'green'), 'app-arch/bzip2-1.0.3-r6:2/foo', Reset(),
            ' ', Color('fg', 'blue'), Bold(), '[1.0.1-r1:0]', Reset())

    def test_dropped_iuse_disabled(self):
        self.formatter.format(FakeOp(
            FakeEbuildSrc('app-arch/bzip2-1.0.3-r6'),
            FakeMutatedPkg('app-arch/bzip2-1.0.3-r6', iuse=['bootstrap']),
        ))
        self.assertOut('[', Color('fg', 'green'), 'ebuild', Reset(),
            '   ', Color('fg', 'yellow'), Bold(), 'R', Reset(), '    ] ',
            Color('fg', 'green'), 'app-arch/bzip2-1.0.3-r6', Reset(),
            ' USE="(', Color('fg', 'yellow'), Bold(), '-bootstrap', Reset(), '%)"')

    def test_dropped_iuse_enabled(self):
        self.formatter.format(FakeOp(
            FakeEbuildSrc('app-arch/bzip2-1.0.3-r6'),
            FakeMutatedPkg('app-arch/bzip2-1.0.3-r6', iuse=['static'], use=['static']),
        ))
        self.assertOut('[', Color('fg', 'green'), 'ebuild', Reset(),
            '   ', Color('fg', 'yellow'), Bold(), 'R', Reset(), '    ] ',
            Color('fg', 'green'), 'app-arch/bzip2-1.0.3-r6', Reset(),
            ' USE="(', Color('fg', 'yellow'), Bold(), '-static', Reset(), '%*)"')

    def test_changed_use(self):
        self.formatter.format(
            FakeOp(FakeEbuildSrc('app-arch/bzip2-1.0.3-r6',
                iuse=['static', 'bootstrap', 'perl', 'foobar', 'rice'],
                use=['static', 'rice']),
            FakeMutatedPkg('app-arch/bzip2-1.0.3-r6',
                iuse=['bootstrap', 'foobar', 'rice', 'kazaam'],
                use=['foobar'])))
        self.assertOut('[', Color('fg', 'green'), 'ebuild', Reset(),
            '   ', Color('fg', 'yellow'), Bold(), 'R', Reset(), '    ] ',
            Color('fg', 'green'), 'app-arch/bzip2-1.0.3-r6', Reset(), ' USE="',
            Color('fg', 'green'), Bold(), 'rice', Reset(), '* ',
            Color('fg', 'yellow'), Bold(), 'static', Reset(), '%* ',
            Color('fg', 'blue'), Bold(), '-bootstrap', Reset(), ' ',
            Color('fg', 'green'), Bold(), '-foobar', Reset(), '* ',
            Color('fg', 'yellow'), Bold(), '-perl', Reset(), '% ',
            '(', Color('fg', 'yellow'), Bold(), '-kazaam', Reset(), '%)"')

    def test_forced_use_verbose(self):
        self.formatter.pkg_get_use = lambda pkg: (set(['static']), set(), set())

        # rebuilt pkg: unchanged static use flag forced on
        self.formatter.format(
            FakeOp(FakeEbuildSrc('app-arch/bzip2-1.0.3-r6', iuse=['static'], use=['static']),
            FakeMutatedPkg('app-arch/bzip2-1.0.3-r6', iuse=['static'], use=['static'])))
        self.assertOut('[', Color('fg', 'green'), 'ebuild', Reset(),
            '   ', Color('fg', 'yellow'), Bold(), 'R', Reset(), '    ] ',
            Color('fg', 'green'), 'app-arch/bzip2-1.0.3-r6', Reset(),
            ' USE="(', Color('fg', 'red'), Bold(), 'static', Reset(), ')"')

    def test_removed_use(self):
        self.formatter.format(
            FakeOp(FakeEbuildSrc('app-arch/bzip2-1.0.3-r6', iuse=['static'], use=['static']),
            FakeMutatedPkg('app-arch/bzip2-1.0.3-r6', iuse=['static', 'bootstrap', 'foo'], use=['static', 'foo'])))
        self.assertOut('[', Color('fg', 'green'), 'ebuild', Reset(),
            '   ', Color('fg', 'yellow'), Bold(), 'R', Reset(), '    ] ',
            Color('fg', 'green'), 'app-arch/bzip2-1.0.3-r6', Reset(),
            ' USE="', Color('fg', 'red'), Bold(), 'static', Reset(), ' ',
            '(', Color('fg', 'yellow'), Bold(), '-bootstrap', Reset(), '%) ',
            '(', Color('fg', 'yellow'), Bold(), '-foo', Reset(), '%*)"')

    def test_end(self):
        self.formatter.format(FakeOp(FakeEbuildSrc('app-arch/bzip2-1.0.3-r6', repo=self.repo1)))
        self.formatter.format(FakeOp(FakeEbuildSrc('app-arch/bzip2-1.0.3-r6', repo=self.repo2)))
        self.fakeout.resetstream()
        self.formatter.end()
        self.assertOut('\nTotal: 2 packages (2 new)\n\n',
            suffix=[''])

class TestPortageVerboseRepoIdFormatter(TestPortageVerboseFormatter):
    suffix = [Color("fg", "cyan"), ' [1]\n']

    def setUp(self):
        TestPortageVerboseFormatter.setUp(self)
        self.repo3 = FakeRepo(
            location='/var/gentoo/repos/repo3', domain_settings=self.domain_settings)

    def newFormatter(self, **kwargs):
        kwargs.setdefault("quiet_repo_display", True)
        return TestPortageVerboseFormatter.newFormatter(self, **kwargs)

    def repo_id(self, repo):
        return ''

    def test_repo_id(self):
        self.formatter.format(FakeOp(FakeEbuildSrc('app-arch/bzip2-1.0.3-r6', repo=self.repo1)))
        self.assertOut('[', Color('fg', 'green'), 'ebuild', Reset(),
            '  ', Color('fg', 'green'), Bold(), 'N', Reset(), '     ] ',
            Color('fg', 'green'), 'app-arch/bzip2-1.0.3-r6', Reset())

    def test_end(self):
        self.formatter.format(FakeOp(FakeEbuildSrc('app-arch/bzip2-1.0.3-r6', repo=self.repo1)))
        self.formatter.format(FakeOp(FakeEbuildSrc('app-arch/bzip2-1.0.3-r6', repo=self.repo2)))
        self.formatter.format(FakeOp(FakeEbuildSrc('app-arch/bzip2-1.0.3-r6', repo=self.repo3)))
        self.fakeout.resetstream()
        self.formatter.end()
        self.assertOut('\nTotal: 3 packages (3 new)\n\n',
            ' ', Color('fg', 'cyan'), '[1]', Reset(),' gentoo (/var/gentoo/repos/gentoo)\n',
            ' ', Color('fg', 'cyan'), '[2]', Reset(),' repo2 (/var/gentoo/repos/repo2)\n',
            ' ', Color('fg', 'cyan'), '[3]', Reset(),' /var/gentoo/repos/repo3\n',
            suffix=[''])



from pkgcore.config import basics
from pkgcore.config.hint import ConfigHint, configurable
from pkgcore.scripts import pebuild
from pkgcore.test.misc import FakePkg, FakeRepo
from pkgcore.test.scripts.helpers import ArgParseMixin
from snakeoil.test import TestCase


class FakeDomain:

    pkgcore_config_type = ConfigHint({'repo': 'ref:repo'}, typename='domain')

    def __init__(self, repo):
        object.__init__(self)
        self.ebuild_repos_unfiltered = repo


@configurable(typename='repo')
def fake_repo():
    pkgs = [
        FakePkg('app-arch/bzip2-1.0.1-r1', slot='0'),
        FakePkg('app-arch/bzip2-1.0.5-r2', slot='0'),
        FakePkg('sys-apps/coreutils-8.25', slot='0'),
        FakePkg('x11-libs/gtk+-2.24', slot='2'),
        FakePkg('x11-libs/gtk+-3.22', slot='3'),
    ]
    repo = FakeRepo(repo_id='gentoo', pkgs=pkgs)
    return repo


domain_config = basics.HardCodedConfigSection({
        'class': FakeDomain,
        'repo': basics.HardCodedConfigSection({'class': fake_repo}),
        'default': True,
        })


class CommandlineTest(TestCase, ArgParseMixin):

    _argparser = pebuild.argparser

    def test_parser(self):
        self.assertError('the following arguments are required: target, phase')
        self.assertError('the following arguments are required: phase', 'dev-util/diffball')

        # working initialization
        config = self.parse('sys-apps/coreutils', 'bar', 'baz', domain=domain_config)
        self.assertEqual(config.phase, ['bar', 'baz'])



"""gentoo/ebuild specific triggers"""

__all__ = (
    "collapse_envd", "string_collapse_envd", "env_update",
    "ConfigProtectInstall", "ConfigProtectUninstall", "preinst_contents_reset",
    "CollisionProtect", "ProtectOwned", "install_into_symdir_protect",
    "InfoRegen", "SFPerms", "FixImageSymlinks", "GenerateTriggers",
)

import fnmatch
import os

from snakeoil.bash import read_bash_dict
from snakeoil.fileutils import AtomicWriteFile
from snakeoil.osutils import listdir_files, normpath, pjoin
from snakeoil.sequences import iflatten_instance, stable_unique

from .. import os_data
from ..fs import livefs
from ..merge import const, errors, triggers
from ..restrictions import values
from ..system import libtool

colon_parsed = frozenset([
    "ADA_INCLUDE_PATH", "ADA_OBJECTS_PATH", "INFODIR", "INFOPATH", "LDPATH",
    "MANPATH", "PATH", "PYTHONPATH", "PKG_CONFIG_PATH", "ROOTPATH"
])

incrementals = frozenset([
    'ADA_INCLUDE_PATH', 'ADA_OBJECTS_PATH', 'CLASSPATH', 'CONFIG_PROTECT',
    'CONFIG_PROTECT_MASK', 'INFODIR', 'INFOPATH', 'KDEDIRS', 'LDPATH',
    'MANPATH', 'PATH', 'PYTHONPATH', 'ROOTPATH', 'PKG_CONFIG_PATH'
])


def collapse_envd(base):
    collapsed_d = {}
    try:
        env_d_files = sorted(listdir_files(base))
    except FileNotFoundError:
        pass
    else:
        for x in env_d_files:
            if x.endswith(".bak") or x.endswith("~") or x.startswith("._cfg") \
                    or len(x) <= 2 or not x[0:2].isdigit():
                continue
            d = read_bash_dict(pjoin(base, x))
            # inefficient, but works.
            for k, v in d.items():
                collapsed_d.setdefault(k, []).append(v)
            del d

    loc_incrementals = set(incrementals)
    loc_colon_parsed = set(colon_parsed)

    # split out env.d defined incrementals..
    # update incrementals *and* colon parsed for colon_separated;
    # incrementals on its own is space separated.

    for x in collapsed_d.pop("COLON_SEPARATED", []):
        v = x.split()
        if v:
            loc_colon_parsed.update(v)

    loc_incrementals.update(loc_colon_parsed)

    # now space.
    for x in collapsed_d.pop("SPACE_SEPARATED", []):
        v = x.split()
        if v:
            loc_incrementals.update(v)

    # now reinterpret.
    for k, v in collapsed_d.items():
        if k not in loc_incrementals:
            collapsed_d[k] = v[-1]
            continue
        if k in loc_colon_parsed:
            collapsed_d[k] = [_f for _f in iflatten_instance(
                x.split(':') for x in v) if _f]
        else:
            collapsed_d[k] = [_f for _f in iflatten_instance(
                x.split() for x in v) if _f]

    return collapsed_d, loc_incrementals, loc_colon_parsed


def string_collapse_envd(envd_dict, incrementals, colon_incrementals):
    """transform a passed in dict to strictly strings"""
    for k, v in envd_dict.items():
        if k not in incrementals:
            continue
        if k in colon_incrementals:
            envd_dict[k] = ':'.join(v)
        else:
            envd_dict[k] = ' '.join(v)


def update_ldso(ld_search_path, offset='/'):
    # we do an atomic rename instead of open and write quick
    # enough (avoid the race iow)
    fp = pjoin(offset, 'etc', 'ld.so.conf')
    new_f = AtomicWriteFile(
        fp, uid=os_data.root_uid, gid=os_data.root_uid, perms=0o644)
    new_f.write("# automatically generated, edit env.d files instead\n")
    new_f.writelines(x.strip()+"\n" for x in ld_search_path)
    new_f.close()


def perform_env_update(root, skip_ldso_update=False):
    d, inc, colon = collapse_envd(pjoin(root, "etc/env.d"))

    l = d.pop("LDPATH", None)
    if l is not None and not skip_ldso_update:
        update_ldso(l, root)

    string_collapse_envd(d, inc, colon)

    new_f = AtomicWriteFile(
        pjoin(root, "etc", "profile.env"),
        uid=os_data.root_uid, gid=os_data.root_gid, perms=0o644)
    new_f.write("# autogenerated.  update env.d instead\n")
    new_f.writelines(f'export {k}="{d[k]}"\n' for k in sorted(d))
    new_f.close()
    new_f = AtomicWriteFile(
        pjoin(root, "etc", "profile.csh"),
        uid=os_data.root_uid, gid=os_data.root_gid, perms=0o644)
    new_f.write("# autogenerated, update env.d instead\n")
    new_f.writelines(f'setenv {k}="{d[k]}"\n' for k in sorted(d))
    new_f.close()


class env_update(triggers.base):

    required_csets = ()
    priority = 5
    _hooks = ('post_unmerge', 'post_merge')

    def trigger(self, engine):
        perform_env_update(engine.offset)


def simple_chksum_compare(x, y):
    found = False
    for k, v in x.chksums.items():
        if k == "size":
            continue
        o = y.chksums.get(k)
        if o is not None:
            if o != v:
                return False
            found = True
    if "size" in x.chksums and "size" in y.chksums:
        return x.chksums["size"] == y.chksums["size"]
    return found


def gen_config_protect_filter(offset, extra_protects=(), extra_disables=()):
    collapsed_d, inc, colon = collapse_envd(pjoin(offset, "etc/env.d"))
    collapsed_d.setdefault("CONFIG_PROTECT", []).extend(extra_protects)
    collapsed_d.setdefault("CONFIG_PROTECT_MASK", []).extend(extra_disables)

    r = [values.StrGlobMatch(normpath(x).rstrip("/") + "/")
         for x in set(stable_unique(collapsed_d["CONFIG_PROTECT"] + ["/etc"]))]
    if len(r) > 1:
        r = values.OrRestriction(*r)
    else:
        r = r[0]
    neg = stable_unique(collapsed_d["CONFIG_PROTECT_MASK"])
    if neg:
        if len(neg) == 1:
            r2 = values.StrGlobMatch(normpath(neg[0]).rstrip("/") + "/",
                                     negate=True)
        else:
            r2 = values.OrRestriction(
                negate=True,
                *[values.StrGlobMatch(normpath(x).rstrip("/") + "/")
                  for x in set(neg)])
        r = values.AndRestriction(r, r2)
    return r


def gen_collision_ignore_filter(offset, extra_ignores=()):
    collapsed_d, inc, colon = collapse_envd(pjoin(offset, "etc/env.d"))
    ignored = collapsed_d.setdefault("COLLISION_IGNORE", [])
    ignored.extend(extra_ignores)
    ignored.extend(["*/.keep", "*/.keep_*"])

    ignored = stable_unique(ignored)
    for i, x in enumerate(ignored):
        if not x.endswith("/*") and os.path.isdir(x):
            ignored[i] = ignored.rstrip("/") + "/*"
    ignored = [values.StrRegex(fnmatch.translate(x))
               for x in stable_unique(ignored)]
    if len(ignored) == 1:
        return ignored[0]
    return values.OrRestriction(*ignored)


class ConfigProtectInstall(triggers.base):

    required_csets = ('install_existing', 'install')
    priority = 100
    _hooks = ('pre_merge',)

    def __init__(self, extra_protects=(), extra_disables=()):
        super().__init__()
        self.renames = {}
        self.extra_protects = extra_protects
        self.extra_disables = extra_disables

    def register(self, engine):
        super().register(engine)
        t2 = ConfigProtectInstall_restore(self.renames)
        t2.register(engine)

    def trigger(self, engine, existing_cset, install_cset):
        # hackish, but it works.
        protected_filter = gen_config_protect_filter(
            engine.offset, self.extra_protects, self.extra_disables).match
        ignore_filter = gen_collision_ignore_filter(engine.offset).match
        protected = {}

        for x in existing_cset.iterfiles():
            if not ignore_filter(x.location) and protected_filter(x.location):
                replacement = install_cset[x]
                if not simple_chksum_compare(replacement, x):
                    protected.setdefault(
                        pjoin(engine.offset,
                              os.path.dirname(x.location).lstrip(os.path.sep)),
                        []).append((os.path.basename(replacement.location),
                                    replacement))

        for dir_loc, entries in protected.items():
            updates = {x[0]: [] for x in entries}
            try:
                existing = sorted(x for x in listdir_files(dir_loc)
                                  if x.startswith("._cfg"))
            except FileNotFoundError:
                # this shouldn't occur.
                continue

            for x in existing:
                try:
                    # ._cfg0000_filename
                    count = int(x[5:9])
                    if x[9] != "_":
                        raise ValueError
                    fn = x[10:]
                except (ValueError, IndexError):
                    continue
                if fn in updates:
                    updates[fn].append((count, fn))

            # now we rename.
            for fname, entry in entries:
                # check for any updates with the same chksums.
                count = 0
                for cfg_count, cfg_fname in updates[fname]:
                    if simple_chksum_compare(livefs.gen_obj(
                            pjoin(dir_loc, cfg_fname)), entry):
                        count = cfg_count
                        break
                    count = max(count, cfg_count + 1)
                try:
                    install_cset.remove(entry)
                except KeyError:
                    # this shouldn't occur...
                    continue
                new_fn = pjoin(dir_loc, "._cfg%04i_%s" % (count, fname))
                new_entry = entry.change_attributes(location=new_fn)
                install_cset.add(new_entry)
                self.renames[new_entry] = entry
            del updates


class ConfigProtectInstall_restore(triggers.base):

    required_csets = ('install',)
    priority = 10
    _hooks = ('post_merge',)

    def __init__(self, renames_dict):
        super().__init__()
        self.renames = renames_dict

    def trigger(self, engine, install_cset):
        for new_entry, old_entry in self.renames.items():
            try:
                install_cset.remove(new_entry)
            except KeyError:
                continue
            install_cset.add(old_entry)
        self.renames.clear()


class ConfigProtectUninstall(triggers.base):

    required_csets = ('uninstall_existing', 'uninstall')
    _hooks = ('pre_unmerge',)

    def trigger(self, engine, existing_cset, uninstall_cset):
        protected_filter = gen_config_protect_filter(engine.offset).match
        ignore_filter = gen_collision_ignore_filter(engine.offset).match

        remove = []
        for x in existing_cset.iterfiles():
            if not ignore_filter(x.location) and protected_filter(x.location):
                recorded_ent = uninstall_cset[x]
                try:
                    if not simple_chksum_compare(recorded_ent, x):
                        # chksum differs.  file stays.
                        remove.append(recorded_ent)
                # If a file doesn't exist we don't need to remove it
                except (FileNotFoundError, NotADirectoryError):
                    pass

        for x in remove:
            del uninstall_cset[x]


class UninstallIgnore(triggers.base):

    required_csets = {
        const.REPLACE_MODE: ('uninstall_existing', 'uninstall', 'old_cset'),
        const.UNINSTALL_MODE: ('uninstall_existing', 'uninstall'),
    }

    _hooks = ('pre_unmerge',)
    _engine_types = triggers.UNINSTALLING_MODES

    def __init__(self, uninstall_ignore=()):
        super().__init__()
        self.uninstall_ignore = uninstall_ignore

    def trigger(self, engine, existing_cset, uninstall_cset, old_cset={}):
        ignore = [values.StrRegex(fnmatch.translate(x), match=True)
                  for x in self.uninstall_ignore]
        ignore_filter = values.OrRestriction(*ignore).match

        remove = [x for x in existing_cset.iterfiles()
                  if ignore_filter(x.location)]
        for x in remove:
            # don't remove matching files being uninstalled
            del uninstall_cset[x]
            # don't remove matching files being replaced
            old_cset.discard(x)


class preinst_contents_reset(triggers.base):

    required_csets = ('new_cset',)
    priority = 1
    _hooks = ('pre_merge',)

    def __init__(self, format_op):
        super().__init__()
        self.format_op = format_op

    def trigger(self, engine, cset):
        # wipe, and get data again; ebuild preinst does untrackable
        # modifications to the fs
        cset.clear()
        cs = engine.new._parent.scan_contents(self.format_op.env["D"])
        if engine.offset != '/':
            cs = cs.insert_offset(engine.offset)
        cset.update(cs)


class FileCollision(triggers.base):
    """Generic livefs file collision trigger."""

    required_csets = {
        const.INSTALL_MODE: ('install', 'install_existing'),
        const.REPLACE_MODE: ('install', 'install_existing', 'old_cset')
    }

    _hooks = ('sanity_check',)
    _engine_types = triggers.INSTALLING_MODES

    suppress_exceptions = False

    def __init__(self, extra_protects=(), extra_disables=(), extra_ignores=()):
        super().__init__()
        self.extra_protects = extra_protects
        self.extra_disables = extra_disables
        self.extra_ignores = extra_ignores

    def collision(self, colliding):
        """Handle livefs file collisions.

        Must be overridden in derived trigger classes.
        """
        raise NotImplementedError

    def trigger(self, engine, install, existing, old_cset=()):
        if not existing:
            return

        # for the moment, we just care about files
        colliding = existing.difference(install.iterdirs())

        # hackish, but it works.
        protected_filter = gen_config_protect_filter(
            engine.offset, self.extra_protects, self.extra_disables).match
        ignore_filter = gen_collision_ignore_filter(
            engine.offset, self.extra_ignores).match

        ignores = []
        for x in colliding:
            if protected_filter(x.location) or ignore_filter(x.location):
                ignores.append(x)

        colliding.difference_update(ignores)
        if not colliding:
            return

        # Wipe the references since we may throw an exception- we don't want
        # potentially millions of references being kept in memory for heavy
        # ignore matches.
        del ignores, protected_filter, ignore_filter
        colliding.difference_update(old_cset)

        if colliding:
            self.collision(colliding)


class CollisionProtect(FileCollision):

    def collision(self, colliding):
        raise errors.BlockModification(
            self, "collision-protect: file(s) already exist: ( %s )" %
            ', '.join(repr(x) for x in sorted(colliding)))


class ProtectOwned(FileCollision):

    def __init__(self, vdb, *args):
        super().__init__(*args)
        self.vdb = vdb

    def collision(self, colliding):
        real_pkgs = (pkg for repo in self.vdb for pkg in repo if pkg.package_is_real)
        collisions = {}

        # TODO: worth parallelizing this vdb scanning?
        for pkg in real_pkgs:
            pkg_file_collisions = pkg.contents.intersection(colliding)
            if pkg_file_collisions:
                collisions[pkg.cpvstr] = pkg_file_collisions

        if collisions:
            pkg_collisions = [
                "( %s ) owned by '%s'" %
                (', '.join(repr(x) for x in sorted(collisions[pkg_cpvstr])), pkg_cpvstr)
                for pkg_cpvstr in sorted(collisions.keys())]
            raise errors.BlockModification(
                self, "protect-owned: %s" % (', '.join(pkg_collisions),))

        # TODO: output a file override warning here


class install_into_symdir_protect(triggers.base):

    required_csets = {
        const.INSTALL_MODE: ('install', 'install_existing'),
        const.REPLACE_MODE: ('install', 'install_existing', 'old_cset')
    }

    _hooks = ('sanity_check',)
    _engine_types = triggers.INSTALLING_MODES

    def __init__(self, extra_protects=(), extra_disables=()):
        super().__init__()
        self.extra_protects = extra_protects
        self.extra_disables = extra_disables

    def trigger(self, engine, install, existing, old_cset=()):
        return
        if not existing:
            return

        # avoid generator madness
        install_into_symdir = []
        for linkset in [install.iterlinks(), existing.iterlinks()]:
            linkset = list(linkset)
            if linkset:
                for inst_file in install.iterfiles():
                    for sym in linkset:
                        if inst_file.location.startswith(sym.location + '/'):
                            install_into_symdir.append(inst_file)

        if install_into_symdir:
            raise errors.BlockModification(
                self,
                "file(s) installed into symlinked dir, will break when "
                "removing files from the original dir: ( %s )" %
                ', '.join(repr(x) for x in sorted(install_into_symdir)))


class InfoRegen(triggers.InfoRegen):

    _label = "ebuild info regen"

    def register(self, engine):
        # wipe pre-existing info triggers.
        for x in self._hooks:
            if x not in engine.hooks:
                continue
            # yucky, but works.
            wipes = [y for y in engine.hooks[x]
                     if y.label == triggers.InfoRegen._label]
            for y in wipes:
                engine.hooks[x].remove(y)
        triggers.InfoRegen.register(self, engine)

    def should_skip_directory(self, basepath, files):
        return any(x.startswith(".keepinfodir")
                   for x in files)

    def trigger(self, engine, *args):
        self.engine = engine
        self.path = pjoin(engine.offset, "etc/env.d")
        triggers.InfoRegen.trigger(self, engine, *args)

    @property
    def locations(self):
        collapsed_d = collapse_envd(self.path)[0]
        l = collapsed_d.get("INFOPATH", ())
        if not l:
            return triggers.InfoRegen.locations
        elif isinstance(l, str):
            l = l.split()
        return l


class SFPerms(triggers.base):
    required_csets = ('new_cset',)
    _hooks = ('pre_merge',)
    _engine_types = triggers.INSTALLING_MODES

    def trigger(self, engine, cset):
        resets = []
        for x in cset.iterfiles():
            if x.mode & 0o4000:
                if x.mode & 0o044:
                    engine.observer.warn(
                        "sfperms: dropping group/world read "
                        f"due to SetGID: {x!r}")
                    resets.append(x.change_attributes(mode=x.mode & ~0o44))
            if x.mode & 0o2000:
                if x.mode & 0o004:
                    engine.observer.warn(
                        f"sfperms: dropping world read due to SetUID: {x!r}")
                    resets.append(x.change_attributes(mode=x.mode & ~0o04))
        cset.update(resets)


def register_multilib_strict_trigger(opts):
    locations = opts.get("MULTILIB_STRICT_DIRS")
    exempt = opts.get(
        "MULTILIB_STRICT_EXEMPT",
        "(perl5|gcc|gcc-lib)")
    deny_pattern = opts.get("MULTILIB_STRICT_DENY")
    if None in (locations, deny_pattern):
        return
    locations = locations.split()
    if not locations:
        return
    elif len(locations) == 1:
        locations = locations[0]
    else:
        locations = f"({'|'.join(locations)})"
    limit_pattern = f"/{locations}/(?!{exempt})"
    while "//" in limit_pattern:
        limit_pattern = limit_pattern.replace("//", "/")
    # this seems semi dodgey, specifically, that it won't catch 'em all
    trig = triggers.BlockFileType(f".*{deny_pattern}.*", limit_pattern)
    return trig


class FixImageSymlinks(triggers.base):

    required_csets = ('new_cset',)
    _hooks = ('pre_merge',)

    def __init__(self, format_op):
        super().__init__()
        self.format_op = format_op

    def trigger(self, engine, cset):
        d = self.format_op.env["D"].rstrip("/") + "/"
        l = [x for x in cset.iterlinks() if x.target.startswith(d)]
        if engine.observer:
            o = engine.observer
            for x in l:
                o.warn(f"correcting {x.location} sym pointing into $D: {x.target}")
        d_len = len(d)

        # drop the leading ${D}, and force an abspath via '/'
        cset.update(
            x.change_attributes(target=pjoin('/', x.target[d_len:]))
            for x in l)


class GenerateTriggers:

    def __init__(self, domain, settings):
        self.domain = domain
        self.opts = {}

        # Pop trigger-related options from the system config, they shouldn't be
        # needed anywhere else and doing so prevents them from leaking into the
        # ebuild env.

        config_lists = (
            "CONFIG_PROTECT", "CONFIG_PROTECT_MASK", "COLLISION_IGNORE",
            "INSTALL_MASK", "UNINSTALL_IGNORE",
        )
        for x in config_lists:
            self.opts[x] = settings.pop(x, [])
            if isinstance(self.opts[x], str):
                self.opts[x] = self.opts[x].split()

        config_opts = (
            "PKGDIR",
            "MULTILIB_STRICT_DIRS", "MULTILIB_STRICT_EXEMPT", "MULTILIB_STRICT_DENY",
            "DEB_REPO_ROOT", "DEB_MAINAINER", "DEB_ARCHITECTURE",
        )
        for x in config_opts:
            self.opts[x] = settings.pop(x, None)

    def __iter__(self):
        yield env_update()

        yield ConfigProtectInstall(
            self.opts["CONFIG_PROTECT"], self.opts["CONFIG_PROTECT_MASK"])
        yield ConfigProtectUninstall()

        if "collision-protect" in self.domain.features:
            yield CollisionProtect(
                self.opts["CONFIG_PROTECT"], self.opts["CONFIG_PROTECT_MASK"],
                self.opts["COLLISION_IGNORE"])

        if "protect-owned" in self.domain.features and "collision-protect" not in self.domain.features:
            yield ProtectOwned(
                self.domain.installed_repos, self.opts["CONFIG_PROTECT"],
                self.opts["CONFIG_PROTECT_MASK"], self.opts["COLLISION_IGNORE"])

        if "multilib-strict" in self.domain.features:
            yield register_multilib_strict_trigger(self.opts)

        if "sfperms" in self.domain.features:
            yield SFPerms()

        yield install_into_symdir_protect(
            self.opts["CONFIG_PROTECT"], self.opts["CONFIG_PROTECT_MASK"])

        # TODO: support multiple binpkg repo targets?
        pkgdir = self.opts.get("PKGDIR", None)
        if pkgdir:
            target_repo = self.domain.binary_repos_raw.get(pkgdir, None)
        else:
            # get the highest priority binpkg repo
            try:
                target_repo = self.domain.binary_repos_raw[0]
            except IndexError:
                target_repo = None
        if target_repo is not None:
            if 'buildpkg' in self.domain.features:
                yield triggers.SavePkg(pristine='no', target_repo=target_repo)
            elif 'pristine-buildpkg' in self.domain.features:
                yield triggers.SavePkg(pristine='yes', target_repo=target_repo)
            elif 'buildsyspkg' in self.domain.features:
                yield triggers.SavePkgIfInPkgset(
                    pristine='yes', target_repo=target_repo, pkgset=self.domain.profile.system)
            elif 'unmerge-backup' in self.domain.features:
                yield triggers.SavePkgUnmerging(target_repo=target_repo)

        if 'splitdebug' in self.domain.features:
            yield triggers.BinaryDebug(mode='split', compress=('compressdebug' in self.domain.features))
        elif 'strip' in self.domain.features or 'nostrip' not in self.domain.features:
            yield triggers.BinaryDebug(mode='strip')

        if '-fixlafiles' not in self.domain.features:
            yield libtool.FixLibtoolArchivesTrigger()

        for x in ("man", "info", "doc"):
            if f"no{x}" in self.domain.features:
                self.opts["INSTALL_MASK"].append(f"/usr/share/{x}")
        l = []
        for x in self.opts["INSTALL_MASK"]:
            x = x.rstrip("/")
            l.append(values.StrRegex(fnmatch.translate(x)))
            l.append(values.StrRegex(fnmatch.translate(f"{x}/*")))
        install_mask = l

        if install_mask:
            if len(install_mask) == 1:
                install_mask = install_mask[0]
            else:
                install_mask = values.OrRestriction(*install_mask)
            yield triggers.PruneFiles(install_mask.match)
            # note that if this wipes all /usr/share/ entries, should
            # wipe the empty dir.

        yield UninstallIgnore(self.opts["UNINSTALL_IGNORE"])
        yield InfoRegen()



"""
Binpkg repository operations for modification, maintenance, etc.

This modules specific operation implementations- said operations are
derivatives of :py:mod:`pkgcore.operations.repo` classes.

Generally speaking you only need to dig through this module if you're trying to
modify a binpkg repository operation- changing how it installs, or changing how
it uninstalls, or adding a new operation (cleaning/cache regen for example).
"""

__all__ = ("install", "uninstall", "replace", "operations")

import os

from snakeoil.compression import compress_data
from snakeoil.klass import steal_docs
from snakeoil.osutils import ensure_dirs, pjoin, unlink_if_exists

from ..fs import tar
from ..log import logger
from ..operations import repo as repo_interfaces
from . import xpak


def discern_loc(base, pkg, extension='.tbz2'):
    return pjoin(base, pkg.category, f"{pkg.package}-{pkg.fullver}{extension}")


_metadata_rewrites = {
    "CONTENTS": "contents",
    "source_repository": "repository",
    "fullslot": "SLOT",
}


def generate_attr_dict(pkg, portage_compatible=True):
    d = {}
    for k in pkg.tracked_attributes:
        if k == "contents":
            continue
        v = getattr(pkg, k)
        if k == 'environment':
            d['environment.bz2'] = compress_data(
                'bzip2', v.bytes_fileobj().read())
            continue
        elif not isinstance(v, str):
            try:
                s = ' '.join(v)
            except TypeError:
                s = str(v)
        else:
            s = v
        d[_metadata_rewrites.get(k, k.upper())] = s
    d[f"{pkg.package}-{pkg.fullver}.ebuild"] = pkg.ebuild.text_fileobj().read()

    # this shouldn't be necessary post portage 2.2.
    # till then, their code requires redundant data,
    # so we've got this.
    if portage_compatible:
        d["CATEGORY"] = pkg.category
        d["PF"] = pkg.PF
    return d


class install(repo_interfaces.install):

    @steal_docs(repo_interfaces.install)
    def add_data(self):
        if self.observer is None:
            end = start = lambda x: None
        else:
            start = self.observer.phase_start
            end = self.observer.phase_end
        pkg = self.new_pkg
        final_path = discern_loc(self.repo.base, pkg, self.repo.extension)
        tmp_path = pjoin(
            os.path.dirname(final_path),
            ".tmp.%i.%s" % (os.getpid(), os.path.basename(final_path)))

        self.tmp_path, self.final_path = tmp_path, final_path

        if not ensure_dirs(os.path.dirname(tmp_path), mode=0o755):
            raise repo_interfaces.Failure(
                f"failed creating directory: {os.path.dirname(tmp_path)!r}")
        try:
            start(f"generating tarball: {tmp_path}")
            tar.write_set(
                pkg.contents, tmp_path, compressor='bzip2',
                parallelize=True)
            end("tarball created", True)
            start("writing Xpak")
            # ok... got a tarball.  now add xpak.
            xpak.Xpak.write_xpak(tmp_path, generate_attr_dict(pkg))
            end("wrote Xpak", True)
            # ok... we tagged the xpak on.
            os.chmod(tmp_path, 0o644)
        except Exception as e:
            try:
                unlink_if_exists(tmp_path)
            except EnvironmentError as e:
                logger.warning(f"failed removing {tmp_path!r}: {e}")
            raise
        return True

    def finalize_data(self):
        os.rename(self.tmp_path, self.final_path)
        return True


class uninstall(repo_interfaces.uninstall):

    @steal_docs(repo_interfaces.uninstall)
    def remove_data(self):
        return True

    @steal_docs(repo_interfaces.uninstall)
    def finalize_data(self):
        os.unlink(discern_loc(self.repo.base, self.old_pkg, self.repo.extension))
        return True


class replace(install, uninstall, repo_interfaces.replace):

    @steal_docs(repo_interfaces.replace)
    def finalize_data(self):
        # we just invoke install finalize_data, since it atomically
        # transfers the new pkg in
        install.finalize_data(self)
        return True


class operations(repo_interfaces.operations):

    def _cmd_implementation_install(self, *args):
        return install(self.repo, *args)

    def _cmd_implementation_uninstall(self, *args):
        return uninstall(self.repo, *args)

    def _cmd_implementation_replace(self, *args):
        return replace(self.repo, *args)



"""
functionality related to downloading files
"""

__all__ = ("fetchable", "mirror", "default_mirror", "uri_list")

from itertools import zip_longest

from snakeoil.klass import generic_equality


class fetchable(metaclass=generic_equality):
    """class representing uri sources for a file and chksum information."""

    __slots__ = ("filename", "uri", "chksums")
    __attr_comparison__ = __slots__

    def __init__(self, filename, uri=None, chksums=None):
        """
        :param filename: filename...
        :param uri: either None (no uri),
            or a sequence of uri where the file is available
        :param chksums: either None (no chksum data),
            or a dict of chksum_type -> value for this file
        """
        self.uri = uri if uri is not None else ()
        self.chksums = chksums if chksums is not None else {}
        self.filename = filename

    def __str__(self):
        chksums = ', '.join(self.chksums)
        return f'({self.filename!r}, {self.uri!r}, {chksums})'

    def __repr__(self):
        return "<%s filename=%r uri=%r chksums=%r @%#8x>" % (
            self.__class__.__name__, self.filename, self.uri, self.chksums,
            id(self))

    def __lt__(self, other):
        return self.filename < other.filename

    def __hash__(self):
        return hash((self.filename, self.uri))

    @property
    def upstream(self):
        """Return a new fetchable with all mirror URIs removed."""
        uri_list = self.uri.remove_mirrors()
        return self.__class__(self.filename, uri=uri_list, chksums=self.chksums)


class mirror(metaclass=generic_equality):
    """uri source representing a mirror tier"""

    __attr_comparison__ = ('mirror_name', 'mirrors')

    __slots__ = ("mirrors", "mirror_name")

    def __init__(self, mirrors, mirror_name):
        """
        :param mirrors: list of hosts that comprise this mirror tier
        :param mirror_name: name of the mirror tier
        """

        if not isinstance(mirrors, tuple):
            mirrors = tuple(mirrors)
        self.mirrors = mirrors
        self.mirror_name = mirror_name

    def __iter__(self):
        return iter(self.mirrors)

    def __str__(self):
        return f"mirror://{self.mirror_name}"

    def __len__(self):
        return len(self.mirrors)

    def __bool__(self):
        return bool(self.mirrors)

    def __getitem__(self, idx):
        return self.mirrors[idx]

    def __repr__(self):
        return f"<{self.__class__} mirror tier={self.mirror_name!r}>"


class unknown_mirror(mirror):
    """Unknown mirror tier."""

    __slots__ = ()

    def __init__(self, mirror_name):
        super().__init__(mirrors=(), mirror_name=mirror_name)


class default_mirror(mirror):

    __slots__ = ()


class uri_list:

    __slots__ = ("_uri_source", "filename", "__weakref__")

    def __init__(self, filename):
        self._uri_source = []
        self.filename = filename

    def add_mirror(self, mirror_inst, sub_uri=None):
        if not isinstance(mirror_inst, mirror):
            raise TypeError("mirror must be a pkgcore.fetch.mirror instance")
        if sub_uri is not None:
            self._uri_source.append((mirror_inst, sub_uri.lstrip('/')))
        else:
            self._uri_source.append(mirror_inst)

    def remove_mirrors(self):
        """Return a new URI source list after dropping all mirror-based URIs."""
        uri_list = self.__class__(self.filename)
        uri_list._uri_source = tuple(x for x in self._uri_source if not isinstance(x, mirror))
        return uri_list

    def add_uri(self, uri):
        self._uri_source.append(uri)

    def finalize(self):
        self._uri_source = tuple(self._uri_source)

    def __iter__(self):
        fname = self.filename
        i = 0
        while i < len(self._uri_source):
            entry = self._uri_source[i]
            if isinstance(entry, str):
                yield entry
            elif isinstance(entry, tuple):
                # TODO: rewrite mirror handling to do this more transparently
                # collect all mirrors at the same priority
                mirrored = []
                while True:
                    m, sub_uri = entry
                    uris = (f"{base_uri.rstrip('/')}/{sub_uri}" for base_uri in m)
                    mirrored.append(uris)
                    try:
                        entry = self._uri_source[i + 1]
                    except IndexError:
                        break
                    if not isinstance(entry, tuple):
                        break
                    i += 1

                # iterate between different mirror groups
                for mirrored_uris in zip_longest(*mirrored):
                    yield from filter(None, mirrored_uris)
            else:
                for base_uri in entry:
                    yield f"{base_uri.rstrip('/')}/{fname}"
            i += 1

    def __str__(self):
        uris = ', '.join(str(x) for x in self._uri_source)
        return f"file: {self.filename}, uri: {uris}"

    def __bool__(self):
        # implemented this way on the off chance an empty sublist is handed in
        for entry in self:
            return True
        return False

    def __len__(self):
        # we do it this way since each item may be a sublist, and to reuse
        # __iter__
        count = 0
        for entry in self:
            count += 1
        return count

    def visit_mirrors(self, invert=False, treat_default_as_mirror=True):
        def is_mirror(item):
            return isinstance(item, mirror) and \
                treat_default_as_mirror == isinstance(item, default_mirror)
        for item in self._uri_source:
            if isinstance(item, tuple):
                if invert != is_mirror(item[0]):
                    yield item
            elif invert != is_mirror(item):
                yield item



from pkgcore.restrictions import boolean, restriction
from snakeoil.test import TestCase

true = restriction.AlwaysBool(node_type='foo', negate=True)
false = restriction.AlwaysBool(node_type='foo', negate=False)


class AlwaysForcableBool(boolean.base):

    __slots__ = ()

    def force_True(self, action, *args):
        yield True

    match = force_False = force_True


class base:

    kls = None

    def test_invalid_restrictions(self):
        self.assertRaises(TypeError, self.kls, 42, node_type='foo')
        base = self.kls(node_type='foo')
        self.assertRaises(TypeError, base.add_restriction, 42)
        self.assertRaises(TypeError, base.add_restriction)

    def test_init_finalize(self):
        final = self.kls(true, node_type='foo', finalize=True)
        # since it becomes a tuple, throws a AttributeError
        self.assertRaises(TypeError, final.add_restriction, false)

        final = self.kls(true, node_type='foo')
        # since it becomes a tuple, throws a AttributeError
        self.assertRaises(TypeError, final.add_restriction, false)

    def test_finalize(self):
        base = self.kls(true, node_type='foo', finalize=False)
        base.add_restriction(false)
        base.finalize()
        self.assertRaises(TypeError, base.add_restriction, true)

    def test_change_restrictions(self):
        base = self.kls(true, false)
        assert self.kls(false, true) == base.change_restrictions(false, true)
        assert self.kls(false, true) != base.change_restrictions(false, true, negate=True)
        assert self.kls(false, true, negate=True) == base.change_restrictions(false, true, negate=True)

    def test_add_restriction(self):
        self.assertRaises(TypeError,
            self.kls(true, finalize=True).add_restriction, false)
        self.assertRaises(TypeError,
            self.kls(node_type='foon').add_restriction, false)
        k = self.kls(finalize=False)
        k.add_restriction(false)
        assert k.restrictions == [false]

    # TODO total_len? what does it do?

class BaseTest(base, TestCase):

    kls = boolean.base

    def test_base(self):
        base = self.kls(true, false, node_type='foo')
        assert len(base) == 2
        assert list(base) == [true, false]
        self.assertRaises(NotImplementedError, base.match, false)
        # TODO is the signature for these correct?
        self.assertRaises(NotImplementedError, base.force_False, false)
        self.assertRaises(NotImplementedError, base.force_True, false)
        self.assertIdentical(base[1], false)


# TODO these tests are way too limited
class AndRestrictionTest(base, TestCase):

    kls = boolean.AndRestriction

    def test_match(self):
        assert self.kls(true, true, node_type='foo').match(None)
        assert not self.kls(false, true, true, node_type='foo').match(None)
        assert not self.kls(true, false, true, node_type='foo').match(None)

    def test_negate_match(self):
        assert self.kls(false, true, node_type='foo', negate=True).match(None)
        assert self.kls(true, false, node_type='foo', negate=True).match(None)
        assert self.kls(false, false, node_type='foo', negate=True).match(None)
        assert not self.kls(true, true, node_type='foo', negate=True).match(None)

    def test_dnf_solutions(self):
        assert self.kls(true, true).dnf_solutions() == [[true, true]]
        assert self.kls(self.kls(true, true), true).dnf_solutions() == [[true, true, true]]
        assert (list(map(set, self.kls(
                    true, true,
                    boolean.OrRestriction(false, true)).dnf_solutions())) ==
            [set([true, true, false]), set([true, true, true])])
        assert self.kls().dnf_solutions() == [[]]

    def test_cnf_solutions(self):
        assert self.kls(true, true).cnf_solutions() == [[true], [true]]
        assert self.kls(self.kls(true, true), true).cnf_solutions() == [[true], [true], [true]]
        assert (list(self.kls(
                    true, true,
                    boolean.OrRestriction(false, true)).cnf_solutions()) ==
            list([[true], [true], [false, true]]))
        assert self.kls().cnf_solutions() == []


class OrRestrictionTest(base, TestCase):

    kls = boolean.OrRestriction

    def test_match(self):
        assert self.kls(true, true, node_type='foo').match(None)
        assert self.kls(false, true, false, node_type='foo').match(None)
        assert self.kls(true, false, false, node_type='foo').match(None)
        assert self.kls(false, false, true, node_type='foo').match(None)
        assert not self.kls(false, false, node_type='foo').match(None)

    def test_negate_match(self):
        for x in ((true, false), (false, true), (true, true)):
            assert not self.kls(node_type='foo', negate=True, *x).match(None)
        assert self.kls(false, false, node_type='foo', negate=True).match(None)

    def test_dnf_solutions(self):
        assert self.kls(true, true).dnf_solutions() == [[true], [true]]
        assert (list(map(set, self.kls(
                    true, true,
                    boolean.AndRestriction(false, true)).dnf_solutions())) ==
            list(map(set, [[true], [true], [false, true]])))
        assert self.kls(self.kls(true, false), true).dnf_solutions() == [[true], [false], [true]]
        assert self.kls().dnf_solutions() == [[]]

    def test_cnf_solutions(self):
        assert self.kls(true, true).cnf_solutions() == [[true, true]]
        assert ([set(x) for x in self.kls(
                    true, true,
                    boolean.AndRestriction(false, true)).cnf_solutions()] ==
            [set(x) for x in [[true, false], [true, true]]])

        assert ([set(x) for x in self.kls(self.kls(
                        true, true,
                        boolean.AndRestriction(false, true))).cnf_solutions()] ==
            [set(x) for x in [[true, false], [true, true]]])

        assert (set(self.kls(
                    self.kls(true, false),
                    true).cnf_solutions()[0]) ==
            set([true, false, true]))

        assert self.kls().cnf_solutions() == []


class JustOneRestrictionTest(base, TestCase):

    kls = boolean.JustOneRestriction

    def test_match(self):
        assert self.kls(true, false, node_type='foo').match(None)
        assert self.kls(false, true, false, node_type='foo').match(None)
        assert not self.kls(false, false,  node_type='foo').match(None)
        assert not self.kls(true, false, true, node_type='foo').match(None)
        assert not self.kls(true, true, true, node_type='foo').match(None)



__all__ = ("LookupFsDev", "ContentsFile")

import os
import stat

from snakeoil import data_source
from snakeoil.chksum import get_handler
from snakeoil.fileutils import AtomicWriteFile, readlines_utf8

from .. import os_data
from ..fs import fs
from ..fs.contents import contentsSet


class LookupFsDev(fs.fsDev):

    __slots__ = ()

    def __init__(self, path, **kwds):
        if any(x not in kwds for x in ("major", "minor", "mode")):
            try:
                st = os.lstat(path)
            except FileNotFoundError:
                st = None
            if st is None or any(f(st.st_mode) for f in
                (stat.S_ISREG, stat.S_ISDIR, stat.S_ISFIFO)):
                kwds["strict"] = True
            else:
                major, minor = fs.get_major_minor(st)
                kwds["major"] = major
                kwds["minor"] = minor
                kwds["mode"] = st.st_mode
        super().__init__(path, **kwds)


class ContentsFile(contentsSet):
    """class wrapping a contents file"""

    def __init__(self, source, mutable=False, create=False):

        if not isinstance(source, (data_source.base, str)):
            raise TypeError("source must be either data_source, or a filepath")
        super().__init__(mutable=True)
        self._source = source

        if not create:
            self.update(self._iter_contents())

        self.mutable = mutable

    def clone(self, empty=False):
        # create is used to block it from reading.
        cset = self.__class__(self._source, mutable=True, create=True)
        if not empty:
            cset.update(self)
        return cset

    def add(self, obj):
        if obj.is_reg:
            # strict checks
            if obj.chksums is None or "md5" not in obj.chksums:
                raise TypeError("fsFile objects need to be strict")

        contentsSet.add(self, obj)

    def _get_fd(self, write=False):
        if isinstance(self._source, str):
            if write:
                return AtomicWriteFile(
                    self._source, uid=os_data.root_uid,
                    gid=os_data.root_gid, perms=0o644)
            return readlines_utf8(self._source, True)
        fobj = self._source.text_fileobj(writable=write)
        if write:
            fobj.seek(0, 0)
            fobj.truncate(0)
        return fobj

    def flush(self):
        return self._write()

    def _iter_contents(self):
        self.clear()
        for line in self._get_fd():
            if not line:
                continue
            s = line.split(" ")
            if s[0] in ("dir", "dev", "fif"):
                path = ' '.join(s[1:])
                if s[0] == 'dir':
                    obj = fs.fsDir(path, strict=False)
                elif s[0] == 'dev':
                    obj = LookupFsDev(path, strict=False)
                else:
                    obj = fs.fsFifo(path, strict=False)
            elif s[0] == "obj":
                path = ' '.join(s[1:-2])
                obj = fs.fsFile(
                    path, chksums={"md5":int(s[-2], 16)},
                        mtime=int(s[-1]), strict=False)
            elif s[0] == "sym":
                try:
                    p = s.index("->")
                    obj = fs.fsLink(' '.join(s[1:p]), ' '.join(s[p+1:-1]),
                        mtime=int(s[-1]), strict=False)

                except ValueError:
                    # XXX throw a corruption error
                    raise
            else:
                raise Exception(f"unknown entry type {line!r}")

            yield obj

    def _write(self):
        md5_handler = get_handler('md5')
        outfile = None
        try:
            outfile = self._get_fd(True)

            for obj in sorted(self):

                if obj.is_reg:
                    s = " ".join(("obj", obj.location,
                        md5_handler.long2str(obj.chksums["md5"]),
                        str(int(obj.mtime))))

                elif obj.is_sym:
                    s = " ".join(("sym", obj.location, "->",
                                   obj.target, str(int(obj.mtime))))

                elif obj.is_dir:
                    s = "dir " + obj.location

                elif obj.is_dev:
                    s = "dev " + obj.location

                elif obj.is_fifo:
                    s = "fif " + obj.location

                else:
                    raise Exception(f"unknown type {type(obj)}: {obj}")
                outfile.write(s + "\n")
            outfile.close()

        finally:
            # if atomic, it forces the update to be wiped.
            del outfile



__all__ = ("hg_syncer",)

import os

from . import base


class hg_syncer(base.VcsSyncer):

    binary = "hg"

    supported_uris = (
        ('hg+', 5),
        ('mercurial+', 5),
    )

    @classmethod
    def is_usable_on_filepath(cls, path):
        hg_path = os.path.join(path, '.hg')
        if cls.disabled or not os.path.isdir(hg_path):
            return None
        return (cls._rewrite_uri_from_stat(hg_path, 'hg+//'),)

    @staticmethod
    def parse_uri(raw_uri):
        if raw_uri.startswith("hg+"):
            return raw_uri[3:]
        elif raw_uri.startswith("mercurial+"):
            return raw_uri[len("mercurial+"):]
        raise base.UriError(
            raw_uri, "doesn't start with hg+ nor mercurial+")

    def _initial_pull(self):
        return [self.binary_path, "clone", self.uri, self.basedir]

    def _update_existing(self):
        # uri may not be set... happens when autodetecting.
        if not self.uri.strip("/"):
            return [self.binary_path, "pull"]
        return [self.binary_path, "pull", "-u", self.uri]



"""make.conf translator.

Converts portage config files into :obj:`pkgcore.config` form.
"""

__all__ = (
    'PortageConfig', 'SecurityUpgradesViaProfile',
)

import configparser
import errno
import os
import sys
from collections import OrderedDict

from snakeoil.bash import read_bash_dict
from snakeoil.compatibility import IGNORED_EXCEPTIONS
from snakeoil.mappings import DictMixin, ImmutableDict
from snakeoil.osutils import access, listdir_files, pjoin

from .. import const
from .. import exceptions as base_errors
from ..config import basics
from ..config import errors as config_errors
from ..config.hint import configurable
from ..fs.livefs import sorted_scan
from ..log import logger
from ..pkgsets.glsa import SecurityUpgrades
from . import const as econst
from . import profiles, repo_objs
from .misc import optimize_incrementals
from .repository import errors as repo_errors


def my_convert_hybrid(manager, val, arg_type):
    """Modified convert_hybrid using a sequence of strings for section_refs."""
    if arg_type.startswith('refs:'):
        subtype = 'ref:' + arg_type.split(':', 1)[1]
        return [basics.LazyNamedSectionRef(manager, subtype, name) for name in val]
    return basics.convert_hybrid(manager, val, arg_type)


@configurable({'ebuild_repo': 'ref:repo', 'vdb': 'ref:repo',
               'profile': 'ref:profile'}, typename='pkgset')
def SecurityUpgradesViaProfile(ebuild_repo, vdb, profile):
    """generate a GLSA vuln. pkgset limited by profile

    Args:
        ebuild_repo (:obj:`pkgcore.ebuild.repository.UnconfiguredTree`): target repo
        vdb (:obj:`pkgcore.repository.prototype.tree`): livefs
        profile (:obj:`pkgcore.ebuild.profiles`): target profile

    Returns:
        pkgset of relevant security upgrades
    """
    arch = profile.arch
    if arch is None:
        raise config_errors.ComplexInstantiationError("arch wasn't set in profiles")
    return SecurityUpgrades(ebuild_repo, vdb, arch)


class ParseConfig(configparser.ConfigParser):
    """Custom ConfigParser class to support returning dict objects."""

    def parse_file(self, f, reset=True):
        """Parse config data from a given file handle.

        By default the underlying config data is reset on each call if it
        exists. This allows multiple files to be easily parsed by a single instance
        without combining all the data in one instance.

        Args:
            f: iterable yielding unicode strings (opened file handle)
            reset (boolean): reset config data if it exists before parsing

        Returns:
            dict: default settings
            dict: regular section settings
        """
        if self._defaults and reset:
            self._defaults = self._dict()
        if self._sections and reset:
            self._sections = self._dict()
        # currently we don't reset section proxies as they should affect
        # this direct data dumping
        self.read_file(f)
        return self._defaults, self._sections


class PortageConfig(DictMixin):
    """Support for portage's config file layout."""

    _supported_repo_types = {}

    def __init__(self, location=None, profile_override=None, **kwargs):
        """
        Args:
            location (optional[str]): path to the portage config directory,
                (defaults to /etc/portage)
            profile_override (optional[str]): profile to use instead of the current system
                profile, i.e. the target of the /etc/portage/make.profile symlink
            root (optional[str]): target root filesystem (defaults to /)
            buildpkg (optional[bool]): forcibly disable/enable building binpkgs, otherwise
                FEATURES=buildpkg from make.conf is used

        Returns:
            dict: config settings
        """
        self._config = {}
        stubconfig = pjoin(const.DATA_PATH, 'stubconfig')

        if location is None:
            path = os.path.abspath(sys.prefix)
            while (parent := os.path.dirname(path)) != path:
                config_root = pjoin(parent, 'etc/portage')
                if os.path.exists(config_root):
                    location = config_root
                    break
                path = parent
            else:
                # fallback to stub config non-Gentoo systems
                location = stubconfig

        # override profile when using stub config
        if location == stubconfig:
            profile_override = pjoin(const.DATA_PATH, 'stubrepo/profiles/default')

        self.dir = location

        # this actually differs from portage parsing- we allow
        # make.globals to provide vars used in make.conf, portage keeps
        # them separate (kind of annoying)
        #
        # this isn't preserving incremental behaviour for features/use unfortunately

        make_conf = {}
        try:
            self.load_make_conf(make_conf, pjoin(const.CONFIG_PATH, 'make.globals'))
        except IGNORED_EXCEPTIONS:
            raise
        except Exception as e:
            raise config_errors.ParsingError("failed to load make.globals") from e
        self.load_make_conf(
            make_conf, pjoin(self.dir, 'make.conf'), required=False,
            allow_sourcing=True, incrementals=True)

        self.root = kwargs.pop('root', make_conf.get("ROOT", "/"))
        gentoo_mirrors = [
            x.rstrip("/") + "/distfiles" for x in make_conf.pop("GENTOO_MIRRORS", "").split()]

        self.features = frozenset(
            optimize_incrementals(make_conf.get('FEATURES', '').split()))

        self._add_sets()
        self._add_profile(profile_override)

        self['vdb'] = basics.AutoConfigSection({
            'class': 'pkgcore.vdb.ondisk.tree',
            'location': pjoin(self.root, 'var', 'db', 'pkg'),
            'cache_location': '/var/cache/edb/dep/var/db/pkg',
        })

        try:
            repos_conf_defaults, repos_conf = self.load_repos_conf(
                pjoin(self.dir, 'repos.conf'))
        except config_errors.ParsingError as e:
            if not getattr(getattr(e, 'exc', None), 'errno', None) == errno.ENOENT:
                raise
            try:
                # fallback to defaults provided by pkgcore
                repos_conf_defaults, repos_conf = self.load_repos_conf(
                    pjoin(const.CONFIG_PATH, 'repos.conf'))
            except IGNORED_EXCEPTIONS:
                raise
            except Exception as e:
                raise config_errors.ParsingError('failed to find a usable repos.conf') from e

        self['ebuild-repo-common'] = basics.AutoConfigSection({
            'class': 'pkgcore.ebuild.repository.tree',
            'default_mirrors': gentoo_mirrors,
            'inherit-only': True,
        })

        repo_map = {}
        repos = []

        for repo_name, repo_opts in list(repos_conf.items()):
            repo_cls = repo_opts.pop('repo-type')
            try:
                repo = repo_cls(
                    self, repo_name=repo_name, repo_opts=repo_opts,
                    repo_map=repo_map, defaults=repos_conf_defaults)
            except repo_errors.UnsupportedRepo as e:
                logger.warning(
                    f'skipping {repo_name!r} repo: unsupported EAPI {str(e.repo.eapi)!r}')
                del repos_conf[repo_name]
                continue

            # only register existent repos
            if os.path.exists(repo_opts['location']):
                self[repo_name] = basics.AutoConfigSection(repo)
                repos.append(repo_name)

        # XXX: Hack for portage-2 profile format support. We need to figure out how
        # to dynamically create this from the config at runtime on attr access.
        profiles.ProfileNode._repo_map = ImmutableDict(repo_map)

        self._make_repo_syncers(repos_conf, make_conf)
        if repos:
            self['repo-stack'] = basics.FakeIncrementalDictConfigSection(
                my_convert_hybrid, {
                    'class': 'pkgcore.repository.multiplex.config_tree',
                    'repos': tuple(repos)})

            self['vuln'] = basics.AutoConfigSection({
                'class': SecurityUpgradesViaProfile,
                'ebuild_repo': 'repo-stack',
                'vdb': 'vdb',
                'profile': 'profile',
            })

        # check if package building was forced on by the user
        forced_buildpkg = kwargs.pop('buildpkg', False)
        if forced_buildpkg:
            make_conf['FEATURES'] += ' buildpkg'

        # finally... domain.
        make_conf.update({
            'class': 'pkgcore.ebuild.domain.domain',
            'repos': tuple(repos),
            'default': True,
            'vdb': ('vdb',),
            'profile': 'profile',
            'root': self.root,
            'config_dir': self.dir,
        })

        self['livefs'] = basics.FakeIncrementalDictConfigSection(
            my_convert_hybrid, make_conf)

    def __setitem__(self, key, value):
        self._config[key] = value

    def __getitem__(self, key):
        return self._config[key]

    def __delitem__(self, key):
        del self._config[key]

    def keys(self):
        return iter(self._config.keys())

    @staticmethod
    def load_make_conf(vars_dict, path, allow_sourcing=False, required=True,
                       allow_recurse=True, incrementals=False):
        """parse make.conf files

        Args:
            vars_dict (dict): dictionary to add parsed variables to
            path (str): path to the make.conf which can be a regular file or
                directory, if a directory is passed all the non-hidden files within
                that directory are parsed in alphabetical order.
        """
        sourcing_command = 'source' if allow_sourcing else None

        if allow_recurse:
            files = sorted_scan(
                os.path.realpath(path), follow_symlinks=True, nonexistent=True,
                hidden=False, backup=False)
        else:
            files = (path,)

        for fp in files:
            try:
                new_vars = read_bash_dict(
                    fp, vars_dict=vars_dict, sourcing_command=sourcing_command)
            except PermissionError as e:
                raise base_errors.PermissionDenied(fp, write=False) from e
            except EnvironmentError as e:
                if e.errno != errno.ENOENT or required:
                    raise config_errors.ParsingError(f"parsing {fp!r}", exception=e) from e
                return

            if incrementals:
                for key in econst.incrementals:
                    if key in vars_dict and key in new_vars:
                        new_vars[key] = f"{vars_dict[key]} {new_vars[key]}"
            # quirk of read_bash_dict; it returns only what was mutated.
            vars_dict.update(new_vars)

    @classmethod
    def load_repos_conf(cls, path):
        """parse repos.conf files

        Args:
            path (str): path to the repos.conf which can be a regular file or
                directory, if a directory is passed all the non-hidden files within
                that directory are parsed in alphabetical order.

        Returns:
            dict: global repo settings
            dict: repo settings
        """
        main_defaults = {}
        repos = {}

        parser = ParseConfig()

        for fp in sorted_scan(
                os.path.realpath(path), follow_symlinks=True, nonexistent=True,
                hidden=False, backup=False):
            try:
                with open(fp) as f:
                    defaults, repo_confs = parser.parse_file(f)
            except PermissionError as e:
                raise base_errors.PermissionDenied(fp, write=False) from e
            except EnvironmentError as e:
                raise config_errors.ParsingError(f"parsing {fp!r}", exception=e) from e
            except configparser.Error as e:
                raise config_errors.ParsingError(f"repos.conf: {fp!r}", exception=e) from e

            if defaults and main_defaults:
                logger.warning(f"repos.conf: parsing {fp!r}: overriding DEFAULT section")
            main_defaults.update(defaults)

            for name, repo_conf in repo_confs.items():
                if name in repos:
                    logger.warning(f"repos.conf: parsing {fp!r}: overriding {name!r} repo")

                # ignore repo if location is unset
                location = repo_conf.get('location', None)
                if location is None:
                    logger.warning(
                        f"repos.conf: parsing {fp!r}: "
                        f"{name!r} repo missing location setting, ignoring repo")
                    continue
                location = os.path.expanduser(location)
                if os.path.isabs(location):
                    repo_conf['location'] = location
                else:
                    # support relative paths based on where repos.conf is located
                    repo_conf['location'] = os.path.abspath(
                        pjoin(os.path.dirname(path), location))

                # repo type defaults to ebuild for compat with portage
                repo_type = repo_conf.get('repo-type', 'ebuild-v1')
                try:
                    repo_conf['repo-type'] = cls._supported_repo_types[repo_type]
                except KeyError:
                    logger.warning(
                        f"repos.conf: parsing {fp!r}: "
                        f"{name!r} repo has unsupported repo-type {repo_type!r}, "
                        "ignoring repo")
                    continue

                # Priority defaults to zero if unset or invalid for ebuild repos
                # while binpkg repos have the lowest priority by default.
                priority = repo_conf.get('priority', None)
                if priority is None:
                    if repo_type.startswith('binpkg'):
                        priority = -10000
                    else:
                        priority = 0

                try:
                    priority = int(priority)
                except ValueError:
                    logger.warning(
                        f"repos.conf: parsing {fp!r}: {name!r} repo has invalid priority "
                        f"setting: {priority!r} (defaulting to 0)")
                    priority = 0
                finally:
                    repo_conf['priority'] = priority

                # register repo
                repos[name] = repo_conf

        if repos:
            # the default repo is gentoo if unset and gentoo exists
            default_repo = main_defaults.get('main-repo', 'gentoo')
            if default_repo not in repos:
                raise config_errors.UserConfigError(
                    f"repos.conf: default repo {default_repo!r} is undefined or invalid")

            if 'main-repo' not in main_defaults:
                main_defaults['main-repo'] = default_repo

            # the default repo has a low priority if unset or zero
            if repos[default_repo]['priority'] == 0:
                repos[default_repo]['priority'] = -1000

        # sort repos via priority, in this case high values map to high priorities
        repos = OrderedDict(
            (k, v) for k, v in
            sorted(repos.items(), key=lambda d: d[1]['priority'], reverse=True))

        return main_defaults, repos

    def _make_repo_syncers(self, repos_conf, make_conf, allow_timestamps=True):
        """generate syncing configs for known repos"""
        rsync_opts = None
        usersync = 'usersync' in self.features

        for repo_name, repo_opts in repos_conf.items():
            d = {'basedir': repo_opts['location'], 'usersync': usersync}

            sync_type = repo_opts.get('sync-type', None)
            sync_uri = repo_opts.get('sync-uri', None)

            if sync_uri:
                # prefix non-native protocols
                if (sync_type is not None and not sync_uri.startswith(sync_type)):
                    sync_uri = f'{sync_type}+{sync_uri}'

                d['uri'] = sync_uri
                d['opts'] = repo_opts.get('sync-opts', '')

                if sync_type == 'rsync':
                    if rsync_opts is None:
                        # various make.conf options used by rsync-based syncers
                        rsync_opts = self._isolate_rsync_opts(make_conf)
                    d.update(rsync_opts)
                    if allow_timestamps:
                        d['class'] = 'pkgcore.sync.rsync.rsync_timestamp_syncer'
                    else:
                        d['class'] = 'pkgcore.sync.rsync.rsync_syncer'
                else:
                    d['class'] = 'pkgcore.sync.base.GenericSyncer'
            elif sync_uri is None:
                # try to autodetect syncing mechanism if sync-uri is missing
                d['class'] = 'pkgcore.sync.base.AutodetectSyncer'
            else:
                # disable syncing if sync-uri is explicitly unset
                d['class'] = 'pkgcore.sync.base.DisabledSync'

            name = 'sync:' + repo_name
            self[name] = basics.AutoConfigSection(d)

    def _add_sets(self):
        self["world"] = basics.AutoConfigSection({
            "class": "pkgcore.pkgsets.filelist.WorldFile",
            "location": pjoin(self.root, econst.WORLD_FILE.lstrip('/'))})
        self["system"] = basics.AutoConfigSection({
            "class": "pkgcore.pkgsets.system.SystemSet",
            "profile": "profile"})
        self["installed"] = basics.AutoConfigSection({
            "class": "pkgcore.pkgsets.installed.Installed",
            "vdb": "vdb"})
        self["versioned-installed"] = basics.AutoConfigSection({
            "class": "pkgcore.pkgsets.installed.VersionedInstalled",
            "vdb": "vdb"})

        set_fp = pjoin(self.dir, "sets")
        try:
            for setname in listdir_files(set_fp):
                # Potential for name clashes here, those will just make
                # the set not show up in config.
                if setname in ("system", "world"):
                    logger.warning(
                        "user defined set %r is disallowed; ignoring",
                        pjoin(set_fp, setname))
                    continue
                self[setname] = basics.AutoConfigSection({
                    "class": "pkgcore.pkgsets.filelist.FileList",
                    "location": pjoin(set_fp, setname)})
        except FileNotFoundError:
            pass

    def _find_profile_path(self, profile_override):
        if profile_override is None:
            make_profile = pjoin(self.dir, 'make.profile')
            if not os.path.islink(make_profile):
                raise config_errors.UserConfigError(f'invalid symlink: {make_profile!r}')
            path = os.path.realpath(make_profile)
        else:
            path = os.path.realpath(profile_override)

        if not os.path.exists(path):
            if profile_override is None:
                raise config_errors.UserConfigError(f'broken symlink: {make_profile!r}')
            else:
                raise config_errors.UserConfigError(f'nonexistent profile: {profile_override!r}')
        return path

    def _add_profile(self, profile_override=None):
        profile = self._find_profile_path(profile_override)
        paths = profiles.OnDiskProfile.split_abspath(profile)
        if paths is None:
            raise config_errors.UserConfigError(
                '%r expands to %r, but no profile detected' %
                (pjoin(self.dir, 'make.profile'), profile))

        user_profile_path = pjoin(self.dir, 'profile')
        if os.path.isdir(user_profile_path):
            self["profile"] = basics.AutoConfigSection({
                "class": "pkgcore.ebuild.profiles.UserProfile",
                "parent_path": paths[0],
                "parent_profile": paths[1],
                "user_path": user_profile_path,
            })
        else:
            self["profile"] = basics.AutoConfigSection({
                "class": "pkgcore.ebuild.profiles.OnDiskProfile",
                "basepath": paths[0],
                "profile": paths[1],
            })

    def _isolate_rsync_opts(self, options):
        """
        pop the misc RSYNC related options littered in make.conf, returning
        a base rsync dict
        """
        base = {}
        opts = []
        extra_opts = []

        opts.extend(options.pop('PORTAGE_RSYNC_OPTS', '').split())
        extra_opts.extend(options.pop('PORTAGE_RSYNC_EXTRA_OPTS', '').split())

        timeout = options.pop('PORTAGE_RSYNC_INITIAL_TIMEOUT', None)
        if timeout is not None:
            base['connection_timeout'] = timeout

        retries = options.pop('PORTAGE_RSYNC_RETRIES', None)
        if retries is not None:
            try:
                retries = int(retries)
                if retries < 0:
                    retries = 10000
                base['retries'] = str(retries)
            except ValueError:
                pass

        proxy = options.pop('RSYNC_PROXY', None)
        if proxy is not None:
            base['proxy'] = proxy.strip()

        if opts:
            base['opts'] = tuple(opts)
        if extra_opts:
            base['extra_opts'] = tuple(extra_opts)

        return base

    def _make_cache(self, cache_format, repo_path):
        """Configure repo cache."""
        # Use md5 cache if it exists or the option is selected, otherwise default
        # to the old flat hash format in /var/cache/edb/dep/*.
        if (os.path.exists(pjoin(repo_path, 'metadata', 'md5-cache')) or
                cache_format == 'md5-dict'):
            kls = 'pkgcore.cache.flat_hash.md5_cache'
            cache_parent_dir = pjoin(repo_path, 'metadata', 'md5-cache')
        else:
            kls = 'pkgcore.cache.flat_hash.database'
            repo_path = pjoin('/var/cache/edb/dep', repo_path.lstrip('/'))
            cache_parent_dir = repo_path

        while not os.path.exists(cache_parent_dir):
            cache_parent_dir = os.path.dirname(cache_parent_dir)
        readonly = (not access(cache_parent_dir, os.W_OK | os.X_OK))

        return basics.AutoConfigSection({
            'class': kls,
            'location': repo_path,
            'readonly': readonly
        })

    def _register_repo_type(supported_repo_types):
        """Decorator to register supported repo types."""
        def _wrap_func(func):
            def wrapped(*args, **kwargs):
                return func(*args, **kwargs)
            name = func.__name__[6:].replace('_', '-')
            supported_repo_types[name] = func
            return wrapped
        return _wrap_func

    @_register_repo_type(_supported_repo_types)
    def _repo_ebuild_v1(self, repo_name, repo_opts, repo_map,
                        defaults, repo_obj=None, repo_dict=None):
        """Create ebuild repo v1 configuration."""
        repo_path = repo_opts['location']

        # XXX: Hack for portage-2 profile format support.
        if repo_obj is None:
            repo_obj = repo_objs.RepoConfig(repo_path, repo_name)
        repo_map[repo_obj.repo_id] = repo_path

        # repo configs
        repo_conf = {
            'class': 'pkgcore.ebuild.repo_objs.RepoConfig',
            'config_name': repo_name,
            'location': repo_path,
            'syncer': 'sync:' + repo_name,
        }
        if repo_dict is not None:
            repo_conf.update(repo_dict)

        # repo trees
        repo = {
            'inherit': ('ebuild-repo-common',),
            'repo_config': 'conf:' + repo_name,
        }

        # metadata cache
        if repo_obj.cache_format is not None:
            cache_name = 'cache:' + repo_name
            self[cache_name] = self._make_cache(repo_obj.cache_format, repo_path)
            repo['cache'] = cache_name

        if repo_name == defaults['main-repo']:
            repo_conf['default'] = True
            repo['default'] = True

        self['conf:' + repo_name] = basics.AutoConfigSection(repo_conf)
        return repo

    @_register_repo_type(_supported_repo_types)
    def _repo_sqfs_v1(self, *args, **kwargs):
        """Create ebuild squashfs repo v1 configuration."""
        repo_name = kwargs['repo_name']
        repo_opts = kwargs['repo_opts']

        repo_path = repo_opts['location']
        sqfs_file = os.path.basename(repo_opts['sync-uri'])
        # XXX: Hack for portage-2 profile format support.
        kwargs['repo_obj'] = repo_objs.SquashfsRepoConfig(sqfs_file, repo_path, repo_name)

        repo_dict = {
            'class': 'pkgcore.ebuild.repo_objs.SquashfsRepoConfig',
            'sqfs_file': sqfs_file,
        }
        kwargs['repo_dict'] = repo_dict
        return self._repo_ebuild_v1(*args, **kwargs)

    @_register_repo_type(_supported_repo_types)
    def _repo_binpkg_v1(self, repo_name, repo_opts, **kwargs):
        """Create binpkg repo v1 configuration."""
        repo = {
            'class': 'pkgcore.binpkg.repository.tree',
            'repo_id': repo_name,
            'location': repo_opts['location'],
        }
        return repo



"""
virtual package
"""

__all__ = ("package", "factory")

from ..restrictions.packages import OrRestriction
from . import metadata


class package(metadata.package):

    """
    Virtual package.

    Mainly useful since it's generating so little attrs on the fly.
    """

    package_is_real = False
    built = True

    __slots__ = ("__dict__",)

    def __init__(self, repo, provider, *a, **kwds):
        super().__init__(repo, *a, **kwds)
        object.__setattr__(self, 'provider', provider)
        object.__setattr__(self, 'data', {})

    def __getattr__(self, key):
        val = None
        if key == "rdepend":
            val = self.provider
        elif key in ("bdepend", "depend", "pdepend", "idepend"):
            val = OrRestriction()
        elif key == "slot":
            val = f"{self.provider.category}-{self.version}"
        else:
            return super().__getattr__(key)
        self.__dict__[key] = val
        return val

    def _fetch_metadata(self):
        data = self._parent._parent_repo._fetch_metadata(self)
        return data


class factory(metadata.factory):
    child_class = package
